import pandas as pd
import numpy as np
from corrai.tsgenerator import DHWaterConsumption, GreyWaterConsumption
from corrai.tsgenerator import Scheduler
import datetime as dt
from pathlib import Path
import pytest

FILES_PATH = Path(__file__).parent / "resources"


class TestDHWaterConsumption:
    def test_get_coefficient_calc_from_period(self):
        df = pd.DataFrame(
            index=pd.date_range("2023-01-01 00:00:00", freq="H", periods=8760)
        )
        start = df.index[0]
        end = df.index[-1]

        # test COSTIC
        dhw1 = DHWaterConsumption(n_dwellings=50)
        dhw1.get_coefficient_calc_from_period(start=start, end=end)
        assert round(dhw1.df_coefficient.loc["2023-04-05 00:00", "coef"], 4) == round(
            1.06 * 0.264 * 1.00, 4
        )
        assert round(dhw1.df_coefficient.loc["2023-08-26 20:00", "coef"], 4) == round(
            1.392 * 0.72 * 1.02, 4
        )
        assert round(dhw1.df_coefficient.loc["2023-09-10 11:00", "coef"], 4) == round(
            1.752 * 0.92 * 1.13, 4
        )

        # #test RE2020
        dhw2 = DHWaterConsumption(n_dwellings=50, method="RE2020")
        dhw2.get_coefficient_calc_from_period(start=start, end=end)
        assert round(dhw2.df_coefficient.loc["2023-04-05 00:00", "coef"], 4) == round(
            0 * 0.95, 4
        )
        assert round(dhw2.df_coefficient.loc["2023-08-25 20:00", "coef"], 4) == round(
            0.022 * 0.95, 4
        )
        assert round(dhw2.df_coefficient.loc["2023-09-10 18:00", "coef"], 4) == round(
            0.011 * 0.95, 4
        )

    def test_get_coefficient_calc_from_period_2(self):
        dhw = DHWaterConsumption(n_dwellings=50)
        start = dt.datetime(2022, 1, 1, 0, 0, 0)
        end = dt.datetime(2024, 10, 20, 1, 0, 0)
        df = dhw.costic_random_shower_distribution(start=start, end=end, seed=42)

        # Check if the sum of random water consumption  distribution is about equal
        # to total estimated consumption
        total_consoECS_COSTIC = dhw.costic_shower_distribution(start=start, end=end)[
            "consoECS_COSTIC"
        ].sum()
        assert np.isclose(
            df["consoECS_COSTIC_random"].sum(), total_consoECS_COSTIC, rtol=0.05
        )

    def test_re2020_shower_distribution(self):
        dhw = DHWaterConsumption(
            n_dwellings=50, s_moy_dwelling=49.6, s_tot_building=2480, method="RE2020"
        )
        start = dt.datetime(2022, 1, 1, 0, 0, 0)
        end = dt.datetime(2024, 10, 20, 1, 0, 0)
        df = dhw.re2020_shower_distribution(start=start, end=end)

        # check Wednesday in April
        assert np.isclose(df.loc["2023-04-05 00:00", "consoECS_RE2020"], 0, rtol=0.05)
        # check Saturday in August
        assert np.isclose(
            df.loc["2023-08-26 20:00", "consoECS_RE2020"], 285.5, rtol=0.05
        )
        # check Sunday in September
        assert np.isclose(
            df.loc["2023-09-10 08:00", "consoECS_RE2020"], 752.7, rtol=0.05
        )

        dhw = DHWaterConsumption(
            n_dwellings=12, s_moy_dwelling=72, s_tot_building=1000, method="RE2020"
        )
        df = dhw.re2020_shower_distribution(start=start, end=end)

        # check Sunday in September
        assert np.isclose(
            df.loc["2023-09-10 08:00", "consoECS_RE2020"], 261.3, rtol=0.05
        )

    def test_scheduler(self):
        schedule_dict = {
            "DAYS": {
                "working_day": {
                    "09:15": {"heating": 17, "extraction_flow_rate": 0},
                    "18:00": {"heating": 21, "extraction_flow_rate": 3000},
                    "19:00": {"heating": 22},
                    "23:00": {"heating": 17, "extraction_flow_rate": 0},
                },
                "Off": {
                    "23:00": {"heating": 17, "extraction_flow_rate": 0},
                },
            },
            "WEEKS": {
                "winter_week": {
                    "Monday": "working_day",
                    "Tuesday": "working_day",
                    "Wednesday": "working_day",
                    "Thursday": "working_day",
                    "Friday": "working_day",
                    "Saturday": "Off",
                    "Sunday": "Off",
                },
                "summer_week": {
                    "Monday": "Off",
                    "Tuesday": "Off",
                    "Wednesday": "Off",
                    "Thursday": "Off",
                    "Friday": "Off",
                    "Saturday": "Off",
                    "Sunday": "Off",
                },
            },
            "PERIODS": [
                (("2009-01-01", "2009-03-31"), "winter_week"),
                (("2009-04-01", "2009-09-30"), "summer_week"),
                (("2009-10-01", "2009-12-31"), "winter_week"),
            ],
            "TZ": "Europe/Paris",
        }

        ref = pd.read_csv(
            filepath_or_buffer=FILES_PATH / "scheduler_test_ref.csv",
            parse_dates=True,
            index_col=0,
        )

        ref.index = pd.date_range(ref.index[0], periods=ref.shape[0], freq="H")

        sched = Scheduler(config_dict=schedule_dict)

        df = sched.get_dataframe(freq="H")

        pd.testing.assert_frame_equal(df, ref)


class TestGreyWaterConsumption:
    def test_get_GWdistribution(self):
        gwc = GreyWaterConsumption(
            n_people=12,
            seed=42,
            v_water_dish=13,
            v_water_clothes=50,
            cycles_clothes_pers=89,  # per year
            cycles_dish_pers=83,  # per year
            duration_dish=4,
            duration_clothes=2,
        )

        df = pd.DataFrame(
            index=pd.date_range("2020-01-01 00:00:00", freq="H", periods=8760)
        )
        start = df.index[0]
        end = df.index[-1]

        gw = gwc.get_GWdistribution(start=start, end=end)

        assert np.isclose(gw["Q_dish"].sum(), 54850.0, rtol=0.05)

        assert np.isclose(gw["Q_washer"].sum(), 3565.25, rtol=0.05)

    def test_warning_error(self):
        with pytest.raises(ValueError):
            GreyWaterConsumption(
                n_people=24,
                dish_washer=False,
                washing_machine=False,
            )



import pandas as pd
import numpy as np
import datetime as dt
from functools import partial
from sklearn.base import TransformerMixin, BaseEstimator
from corrai.math import time_gradient
from scipy.ndimage import gaussian_filter1d
from abc import ABC, abstractmethod


class PdTransformerBC(TransformerMixin, BaseEstimator, ABC):
    def __init__(self):
        self.columns = None
        self.index = None

    def get_feature_names_out(self, input_features=None):
        return self.columns

    @abstractmethod
    def fit(self, X, y=None):
        """Operations happening during fitting process"""
        pass

    @abstractmethod
    def transform(self, X):
        """Operations happening during transforming process"""
        pass


class PdIdentity(PdTransformerBC):
    """
    A custom transformer that returns the input data without any modifications.

    This transformer is useful when you want to include an identity transformation step
    in a scikit-learn pipeline, where the input data should be returned unchanged.

    Parameters:
    -----------
    None

    Methods:
    --------
    fit(X, y=None):
        This method does nothing and simply returns the transformer instance.

        Parameters:
        -----------
        X : array-like, shape (n_samples, n_features)
            The input data.

        y : array-like, shape (n_samples,), optional (default=None)
            The target values.

        Returns:
        --------
        self : object
            The transformer instance itself.

    transform(X):
        This method returns the input data without any modifications.

        Parameters:
        -----------
        X : array-like, shape (n_samples, n_features)
            The input data.

        Returns:
        --------
        transformed_X : array-like, shape (n_samples, n_features)
            The input data without any modifications.
    """

    def __init__(self):
        super().__init__()

    def fit(self, X, y=None):
        self.columns = X.columns
        self.index = X.index
        return self

    def transform(self, X):
        return X


class PdDropna(PdTransformerBC):
    """A class to drop NaN values in a Pandas DataFrame.

    Parameters
    ----------
    how : str, default 'all'
        How to drop missing values in the data. 'all' drops the row/column if
        all the values are missing, 'any' drops the row/column if any value is
        missing, and a number 'n' drops the row/column if there are at least
        'n' missing values.

    Attributes
    ----------
    how : str
        How to drop missing values in the data.

    Methods
    -------
    fit(X, y=None)
        Returns self.

    transform(X)
        Drops the NaN values in the Pandas DataFrame `X` based on the `how`
        attribute.
        Returns the DataFrame with the NaN values dropped.
    """

    def __init__(self, how="all"):
        super().__init__()
        self.how = how

    def fit(self, X, y=None):
        self.columns = X.columns
        self.index = X.index
        return self

    def transform(self, X):
        return X.dropna(how=self.how)


class PdRenameColumns(PdTransformerBC):
    """
    Scikit-learn transformer that renames columns of a Pandas DataFrame.

    Parameters
    ----------
    new_names: list or dict
        A list or a dictionary of new names for columns of a DataFrame.
        If it is a list, it must have the same length as the number of columns
        in the DataFrame. If it is a dictionary, keys must be the old names of
        columns and values must be the new names.

    Attributes
    ----------
    new_names: list or dict
        A list or a dictionary of new names for columns of a DataFrame.

    Methods
    -------
    fit(self, x, y=None)
       No learning is performed, the method simply returns self.

    transform(self, x)
        Renames columns of a DataFrame.

    inverse_transform(self, x)
        Renames columns of a DataFrame.
    """

    def __init__(self, new_names):
        super().__init__()
        self.new_names = new_names

    def fit(self, X, y=None):
        self.columns = X.columns
        self.index = X.index
        return self

    def transform(self, X):
        X.columns = self.new_names
        self.columns = X.columns
        return X

    def inverse_transform(self, X):
        if not isinstance(X, pd.DataFrame):
            X = pd.DataFrame(X)

        X.columns = self.new_names
        self.columns = X.columns
        return X


class PdSkTransformer(PdTransformerBC):
    """A transformer class to apply scikit transformers on a pandas DataFrame

    This class takes in a scikit-learn transformers as input and applies the
    transformer to a pandas DataFrame. The resulting data will be a pandas
    DataFrame with the same index and columns as the input DataFrame.

    Parameters
    ----------
    transformer : object
        A scikit-learn transformer to apply on the data.

    Attributes
    ----------
    transformer : object
        A scikit-learn transformer that is fitted on the data.

    Methods
    -------
    fit(x, y=None)
        Fit the scaler to the input data `x`

    transform(x)
        Apply the transformer to the input data `x` and return the result
        as a pandas DataFrame.

    inverse_transform(x)
        Apply the inverse transformer to the input data `x` and return the
        result as a pandas DataFrame.

    """

    def __init__(self, transformer):
        super().__init__()
        self.transformer = transformer

    def fit(self, X, y=None):
        self.transformer.fit(X)
        self.columns = X.columns
        self.index = X.index
        return self

    def transform(self, X):
        return pd.DataFrame(
            data=self.transformer.transform(X), index=X.index, columns=X.columns
        )

    def inverse_transform(self, X):
        if isinstance(X, np.ndarray):
            X = pd.DataFrame(X)
            X.columns = self.columns
        return pd.DataFrame(
            data=self.transformer.inverse_transform(X), index=X.index, columns=X.columns
        )


class PdDropThreshold(PdTransformerBC):
    """Class replacing values in a pandas DataFrame by NaN based on
    threshold values.

    This class implements the scikit-learn transformer API and can be used in
    a scikit-learn pipeline.

    Parameters
    ----------
    upper : float, optional (default=None)
        The upper threshold for values in the DataFrame. Values greater than
        The upper threshold for values in the DataFrame. Values greater than
        this threshold will be replaced.
    lower : float, optional (default=None)
        The lower threshold for values in the DataFrame. Values less than
        this threshold will be replaced.

    Attributes
    ----------
    lower : float
        The lower threshold for values in the DataFrame.
    upper : float
        The upper threshold for values in the DataFrame.

    Methods
    -------
    fit(self, X, y=None)
        No learning is performed, the method simply returns self.

    transform(self, X)
        Transforms the input DataFrame by replacing values based on the
        specified upper and lower thresholds.
    """

    def __init__(self, upper=None, lower=None):
        super().__init__()
        self.lower = lower
        self.upper = upper

    def fit(self, X, y=None):
        self.columns = X.columns
        self.index = X.index
        return self

    def transform(self, X):
        if self.lower is not None:
            lower_mask = X < self.lower
        else:
            lower_mask = pd.DataFrame(
                np.full(X.shape, False), index=X.index, columns=X.columns
            )

        if self.upper is not None:
            upper_mask = X > self.upper
        else:
            upper_mask = pd.DataFrame(
                np.full(X.shape, False), index=X.index, columns=X.columns
            )

        X[np.logical_or(lower_mask, upper_mask)] = np.nan

        return X


class PdDropTimeGradient(PdTransformerBC):
    """
    A transformer that removes values in a DataFrame based on the time gradient.

    The time gradient is calculated as the difference of consecutive values in
    the time series divided by the time delta between each value.
    If the gradient is below the `lower_rate` or above the `upper_rate`,
    then the value is set to NaN.

    Parameters
    ----------
    dropna : bool, default=True
        Whether to remove NaN values from the DataFrame before processing.
    upper_rate : float, optional
        The upper rate threshold. If the gradient is greater than or equal to
        this value, the value will be set to NaN.
    lower_rate : float, optional
        The lower rate threshold. If the gradient is less than or equal to
         this value, the value will be set to NaN.

    Attributes
    ----------
    None

    Methods
    -------
    fit(X, y=None)
        No learning is performed, the method simply returns self.
    transform(X)
        Removes values in the DataFrame based on the time gradient.

    Returns
    -------
    DataFrame
        The transformed DataFrame.
    """

    def __init__(self, dropna=True, upper_rate=None, lower_rate=None):
        super().__init__()
        self.dropna = dropna
        self.upper_rate = upper_rate
        self.lower_rate = lower_rate

    def fit(self, X, y=None):
        self.columns = X.columns
        self.index = X.index
        return self

    def transform(self, X):
        X_transformed = []
        for column in X.columns:
            X_column = X[column]
            if self.dropna:
                original_index = X_column.index.copy()
                X_column = X_column.dropna()

            time_delta = X_column.index.to_series().diff().dt.total_seconds()
            abs_der = abs(X_column.diff().divide(time_delta, axis=0))
            abs_der_two = abs(X_column.diff(periods=2).divide(time_delta, axis=0))
            if self.upper_rate is not None:
                mask_der = abs_der >= self.upper_rate
                mask_der_two = abs_der_two >= self.upper_rate
            else:
                mask_der = pd.Series(
                    np.full(X_column.shape, False),
                    index=X_column.index,
                    name=X_column.name,
                )
                mask_der_two = mask_der

            if self.lower_rate is not None:
                mask_constant = abs_der <= self.lower_rate
            else:
                mask_constant = pd.Series(
                    np.full(X_column.shape, False),
                    index=X_column.index,
                    name=X_column.name,
                )

            mask_to_remove = np.logical_and(mask_der, mask_der_two)
            mask_to_remove = np.logical_or(mask_to_remove, mask_constant)

            X_column[mask_to_remove] = np.nan
            if self.dropna:
                X_column = X_column.reindex(original_index)
            X_transformed.append(X_column)
        return pd.concat(X_transformed, axis=1)


class PdApplyExpression(PdTransformerBC):
    """A transformer class to apply a mathematical expression on a Pandas
    DataFrame.

    This class implements a transformer that can be used to apply a
     mathematical expression to a Pandas DataFrame.
    The expression can be any valid Python expression that
    can be evaluated using the `eval` function.

    Parameters
    ----------
    expression : str
        A string representing a valid Python expression.
        The expression can use any variables defined in the local scope,
        including the `X` variable that is passed to the `transform` method
         as the input data.

    Attributes
    ----------
    expression : str
        The mathematical expression that will be applied to the input data.

    """

    def __init__(self, expression):
        super().__init__()
        self.expression = expression

    def fit(self, X, y=None):
        self.columns = X.columns
        self.index = X.index
        return self

    def transform(self, X):
        return eval(self.expression)


class PdTimeGradient(PdTransformerBC):
    """
    A class to calculate the time gradient of a pandas DataFrame,
     which is the derivative of the data with respect to time.

    Parameters
    ----------
    dropna : bool, optional (default=True)
        Whether to drop NaN values before calculating the time gradient.

    Attributes
    ----------
    dropna : bool
        The dropna attribute of the class.

    Methods
    -------
    fit(X, y=None)
        Fits the transformer to the data. Does not modify the input data.

    transform(X)
        Transforms the input data by calculating the time gradient of
         the data.

    """

    def __init__(self):
        super().__init__()

    def fit(self, X, y=None):
        self.columns = X.columns
        self.index = X.index
        return self

    def transform(self, X):
        original_index = X.index.copy()
        derivative = time_gradient(X)
        return derivative.reindex(original_index)


class PdFillNa(PdTransformerBC):
    """
    A class that extends scikit-learn's TransformerMixin and BaseEstimator
    to fill missing values in a Pandas DataFrame.

    Parameters:
        value: scalar, dict, Series, or DataFrame
            Value(s) used to replace missing values.
        method: str, optional
            Interpolation method used to fill missing values.
            Must be one of 'backfill', 'bfill', 'pad', 'ffill', 'nearest',
            'interpolate' or None.
            If None, the missing values are left as is.


    Methods:
        fit(self, X, y=None):
            Does nothing. Returns the object itself.
        transform(self, X):
            Fill missing values in the input DataFrame.
    """

    def __init__(self, value=None, method=None):
        super().__init__()
        self.value = value
        self.method = method

    def fit(self, X, y=None):
        self.columns = X.columns
        self.index = X.index
        return self

    def transform(self, X):
        return X.fillna(value=self.value, method=self.method)


class PdResampler(PdTransformerBC):
    """A transformer class that resamples a Pandas DataFrame

    This class performs resampling on a Pandas DataFrame using the specified
    `rule` and `method`. The resampling operation aggregates the data by time
    intervals of the given `rule` and applies the specified `method` to the
    data within each interval.

    This transformer extends the scikit-learn `TransformerMixin` and `BaseEstimator`
    classes.

    Parameters:
    -----------
    rule : str
        The frequency of resampling. The `rule` must be a string that can be
        passed to the `resample()` method of a Pandas DataFrame.
        It can also be a datetime.timedelta object
    method : str, callable, or list of str/callable, default None
        The aggregation method to be used in resampling. The `method`
        must be a string or callable that can be passed to the `agg()` method
        of a resampled Pandas DataFrame.

    Attributes:
    ----------
    columns : array-like
        The names of the columns in the input data.
    resampled_index : DatetimeIndex
        The index of the resampled Pandas DataFrame.

    Methods:
    -------
    get_feature_names_out(input_features=None)
        Returns the names of the output features.
    fit(X, y=None)
        Fit the transformer to the input data.
    transform(X)
        Perform the resampling operation on the input data and return the
        resampled DataFrame.

    Returns:
    -------
    X_resampled : Pandas DataFrame
        The resampled Pandas DataFrame.
    """

    def __init__(self, rule, method=None):
        super().__init__()
        self.rule = rule
        self.method = method
        self.resampled_index = None

    def fit(self, X, y=None):
        self.columns = X.columns
        self.resampled_index = X.resample(self.rule).asfreq()
        return self

    def transform(self, X):
        X = X.apply(pd.to_numeric)
        X_resampled = X.resample(self.rule).agg(self.method)
        self.resampled_index = X_resampled.index
        return X_resampled


class PdInterpolate(PdTransformerBC):
    """A class that implements interpolation of missing values in
     a Pandas DataFrame.

    This class is a transformer that performs interpolation of missing
    values in a Pandas DataFrame, using the specified `method`.

    Parameters:
    -----------
    method : str or None, default None
        The interpolation method to use. If None, the default interpolation
         method of the Pandas DataFrame `interpolate()` method will be used.

    Attributes:
    -----------
    columns : Index or None
        The columns of the input DataFrame. Will be set during fitting.
    index : Index or None
        The index of the input DataFrame. Will be set during fitting.

    Methods:
    --------
    fit(X, y=None):
        Fit the transformer to the input DataFrame X. This method will set
         the `columns` and `index` attributes of the transformer,
          and return the transformer instance.
    transform(X):
        Transform the input DataFrame X by performing interpolation of
         missing values using the
        specified `method`. Returns the transformed DataFrame.

    Returns:
    -------
    A transformed Pandas DataFrame with interpolated missing values.
    """

    def __init__(self, method=None):
        super().__init__()
        self.method = method

    def fit(self, X, y=None):
        self.columns = X.columns
        self.index = X.index
        return self

    def transform(self, X):
        return X.interpolate(method=self.method)


class PdColumnResampler(PdTransformerBC):
    """Resample time series data in a pandas DataFrame based on different
    resampling methods for different columns.

    WARNING Use PdResampler if you want to resample all the columns with
    the same aggregation method

    Parameters
    ----------
    rule : str
        The offset string or object representing the target resampling
        frequency.
    columns_method : list of tuples
        List of tuples containing a list of column names and an associated
        resampling method. The method should be a callable that can be passed
        to the `agg()` method of a pandas DataFrame.
    remainder : str or callable, default='drop'
        If 'drop', drop the non-aggregated columns from the resampled output.
        If callable, must be a function that takes a DataFrame as input and
        returns a DataFrame. Any non-aggregated columns not included in the
        output of this function will be dropped from the resampled output.

    Attributes
    ----------
    resampled_index : pandas.DatetimeIndex or None
        Index of the resampled data after fitting.
    columns : list of str or None
        List of columns used to fit the transformer. If `remainder` is set to
        'drop', this will only contain the columns used for aggregation.
        Otherwise, it will contain all columns in the original DataFrame.


    Methods
    -------
    fit(X, y=None)
        Fit the transformer to the input DataFrame.
    transform(X)
        Resample the input DataFrame based on the specified resampling methods.
    get_feature_names_out()
        Return the names of the new features created in `transform()`.

    """

    def __init__(self, rule, columns_method, remainder="drop"):
        super().__init__()
        self.rule = rule
        self.columns_method = columns_method
        self.resampled_index = None

        if remainder != "drop" and not callable(remainder):
            raise ValueError(
                "If remainder is no set to drop. Aggregation "
                "method must be provided to ensure transformed "
                "index consistency"
            )
        else:
            self.remainder = remainder
        self._check_columns_method()

    def _check_columns_method(self):
        if not isinstance(self.columns_method, list):
            raise ValueError(
                "Columns_method must be a list of Tuple"
                "first index shall be a list of columns names,"
                "second index shall be an aggregation method"
            )
        for elmt in self.columns_method:
            if not isinstance(elmt[0], list):
                raise ValueError("Tuple first element must be a list" "of columns")
            if not callable(elmt[1]):
                raise ValueError("Tuple second element must be a" "callable")

    def _check_columns(self, X):
        for col_list, _ in self.columns_method:
            for col in col_list:
                if col not in X.columns:
                    raise ValueError("Columns in columns_method not found in" "X")

    def fit(self, X, y=None):
        self._check_columns(X)
        if self.remainder == "drop":
            self.columns = []
            for col_list, _ in self.columns_method:
                self.columns += col_list
        else:
            self.columns = X.columns

        return self

    def transform(self, X):
        if self.columns_method:
            transformed_X = pd.concat(
                [
                    X[tup[0]].resample(self.rule).agg(tup[1])
                    for tup in self.columns_method
                ],
                axis=1,
            )
        else:
            transformed_X = pd.DataFrame()

        if self.remainder != "drop":
            remaining_col = [
                col for col in X.columns if col not in transformed_X.columns
            ]
            transformed_X = pd.concat(
                [
                    transformed_X,
                    X[remaining_col].resample(self.rule).agg(self.remainder),
                ],
                axis=1,
            )

        return transformed_X[self.columns]


class PdAddTimeLag(PdTransformerBC):
    """
     PdAddTimeLag - A transformer that adds lagged features to a pandas
     DataFrame.

    This transformer creates new features based on the provided features
    lagged by the given time lag.

    Parameters:
    -----------
    time_lag : datetime.timedelta
        The time lag used to shift the provided features. A positive time lag
        indicates that the new features will contain information from the past,
         while a negative time lag indicates that the new features will
        contain information from the future.

    features_to_lag : list or None, optional (default=None)
        The list of feature names to lag. If None, all features in the input
         DataFrame will be lagged.

    custom_feature_marker : str or None, optional (default=None)
        The string used to prefix the names of the new lagged features.
        If None, the feature names will be prefixed with the string
        representation of the `time_lag` parameter followed by an underscore.

    drop_resulting_nan : bool, optional (default=False)
        Whether to drop rows with NaN values resulting from the lag operation.

    """

    def __init__(
        self,
        time_lag,
        features_to_lag=None,
        feature_marker=None,
        drop_resulting_nan=False,
    ):
        super().__init__()
        if not isinstance(time_lag, dt.timedelta):
            raise ValueError(
                "Invalid time_lag value. You must provide "
                "a datetime.timedelta object"
            )
        self.time_lag = time_lag
        self.features_to_lag = features_to_lag
        self.drop_resulting_nan = drop_resulting_nan

        if feature_marker is None:
            self.feature_marker = str(time_lag) + "_"
        else:
            if not isinstance(feature_marker, str):
                raise ValueError("feature_marker must be a string")
            self.feature_marker = feature_marker

    def fit(self, X, y=None):
        return self

    def transform(self, X):
        if self.features_to_lag is None:
            self.features_to_lag = X.columns
        to_lag = X[self.features_to_lag].copy()
        to_lag.index = to_lag.index + self.time_lag
        to_lag.columns = self.feature_marker + to_lag.columns
        X_transformed = pd.concat([X, to_lag], axis=1)
        if self.drop_resulting_nan:
            X_transformed = X_transformed.dropna()
        return X_transformed


class PdGaussianFilter1D(PdTransformerBC):
    """
    A transformer that applies a 1D Gaussian filter to a Pandas DataFrame.
    The Gaussian filter is a widely used smoothing filter that effectively
    reduces the high-frequency noise in an input signal.

    Parameters
    ----------
    sigma : float, default=5
        Standard deviation of the Gaussian kernel.
        In practice, the value of sigma determines the level of smoothing
        applied to the input signal. A larger value of sigma results in a
         smoother output signal, while a smaller value results in less
          smoothing. However, too large of a sigma value can result in the
           loss of important features or details in the input signal.

    mode : str, default='nearest'
        Points outside the boundaries of the input are filled according to
        the given mode. The default, 'nearest' mode is used to set the values
        beyond the edge of the array equal to the nearest edge value.
        This avoids introducing new values into the smoothed signal that
        could bias the result. Using 'nearest' mode can be particularly useful
        when smoothing a signal with a known range or limits, such as a time
        series with a fixed start and end time.

    truncate : float, default=4.
        The filter will ignore values outside the range
        (mean - truncate * sigma) to (mean + truncate * sigma).
        The truncate parameter is used to define the length of the filter
        kernel, which determines the degree of smoothing applied to the input
        signal.

    Attributes
    ----------
    columns : list
        The column names of the input DataFrame.
    index : pandas.Index
        The index of the input DataFrame.

    Methods
    -------
    get_feature_names_out(input_features=None)
        Get output feature names for the transformed data.
    fit(X, y=None)
        Fit the transformer to the input data.
    transform(X, y=None)
        Transform the input data by applying the 1D Gaussian filter.

    """

    def __init__(self, sigma=5, mode="nearest", truncate=4.0):
        super().__init__()
        self.sigma = sigma
        self.mode = mode
        self.truncate = truncate

    def fit(self, X, y=None):
        self.columns = X.columns
        self.index = X.index
        return self

    def transform(self, X, y=None):
        if self.columns is None:
            raise ValueError(
                "Transformer is not fitted yet. Perform fitting using fit() method"
            )

        gauss_filter = partial(
            gaussian_filter1d, sigma=self.sigma, mode=self.mode, truncate=self.truncate
        )

        return X.apply(gauss_filter)


class PdCombineColumns(PdTransformerBC):
    """
    A class that combines multiple columns in a pandas DataFrame using a specified
    function.

    Parameters
    ----------
        columns_to_combine (list or None): A list of column names to combine.
            If None, all columns will be combined.
        function (callable or None): A function or method to apply for combining
            columns.
        function_kwargs (dict or None): Additional keyword arguments to pass to the
            combining function.
        drop_columns (bool): If True, the original columns to combine will be dropped
            from the DataFrame. If False, the original columns will be retained.
        label_name (str): The name of the new column that will store the combined
            values.

    Attributes
    ----------
        columns : list
            The column names of the input DataFrame.
        index : pandas.Index
            The index of the input DataFrame.

    Methods
    -------
        get_feature_names_out(input_features=None)
            Get output feature names for the transformed data.
        fit(X, y=None)
            Fit the transformer to the input data.
        transform(X, y=None)
            Transform the input data by applying the function
    """

    def __init__(
        self,
        columns_to_combine=None,
        function=None,
        function_kwargs=None,
        drop_columns=False,
        label_name="combined",
    ):
        super().__init__()
        if function_kwargs is None:
            function_kwargs = {}
        self.function_kwargs = function_kwargs
        self.columns_to_combine = columns_to_combine
        self.function = function
        self.drop_columns = drop_columns
        self.label_name = label_name

    def fit(self, X, y=None):
        self.columns = X.columns
        self.index = X.index
        for lab in self.columns_to_combine:
            if lab not in self.columns:
                raise ValueError(f"{lab} is not found in X DataFrame columns")
        return self

    def transform(self, X):
        X_transformed = X.copy()
        if self.drop_columns:
            col_to_return = [
                col for col in self.columns if col not in self.columns_to_combine
            ]
        else:
            col_to_return = list(self.columns)

        X_transformed[self.label_name] = self.function(
            X_transformed[self.columns_to_combine], **self.function_kwargs
        )

        col_to_return.append(self.label_name)

        return X_transformed[col_to_return]



import numpy as np
import pandas as pd
import corrai.custom_transformers as ct
import json
from collections import defaultdict
from corrai.utils import check_datetime_index
from corrai.custom_transformers import PdIdentity
import plotly.graph_objects as go
import datetime as dt
from sklearn.compose import ColumnTransformer
from sklearn.pipeline import make_pipeline

TRANSFORMER_MAP = {
    "dropna": ct.PdDropna,
    "rename_columns": ct.PdRenameColumns,
    "sk_transformer": ct.PdSkTransformer,
    "drop_threshold": ct.PdDropThreshold,
    "drop_time_gradient": ct.PdDropTimeGradient,
    "apply_expression": ct.PdApplyExpression,
    "time_gradient": ct.PdTimeGradient,
    "fill_na": ct.PdFillNa,
    "resample": ct.PdResampler,
    "interpolate": ct.PdInterpolate,
    "gaussian_filter": ct.PdGaussianFilter1D,
}

RESAMPLE_METHS = {"mean": np.mean, "sum": np.sum}

COLOR_PALETTE = ["#FFAD85", "#FF8D70", "#ED665A", "#52E0B6", "#479A91"]


def get_transformers_keys():
    return TRANSFORMER_MAP.keys()


def missing_values_dict(df):
    return {
        "Number_of_missing": df.count(),
        "Percent_of_missing": (1 - df.count() / df.shape[0]) * 100,
    }


def set_multi_yaxis_layout(figure, ax_dict, axis_space):
    nb_right_y_axis = len(set(ax_dict.values()))
    x_right_space = 1 - axis_space * (nb_right_y_axis - 1)
    figure.update_xaxes(domain=(0, x_right_space))
    ax_args = {
        f"yaxis{2 + i}": dict(position=x_right_space + i * axis_space)
        for i in range(nb_right_y_axis)
    }
    figure.update_layout(**ax_args)


def interpolate_color(color1, color2, t):
    """Interpolate between two colors based on a parameter t"""
    r = int((1 - t) * int(color1[1:3], 16) + t * int(color2[1:3], 16))
    g = int((1 - t) * int(color1[3:5], 16) + t * int(color2[3:5], 16))
    b = int((1 - t) * int(color1[5:7], 16) + t * int(color2[5:7], 16))
    return f"#{r:02x}{g:02x}{b:02x}"


def darken_color(color, factor):
    r, g, b = int(color[1:3], 16), int(color[3:5], 16), int(color[5:7], 16)
    r = int(r * factor)
    g = int(g * factor)
    b = int(b * factor)
    darkened_color = f"#{r:02X}{g:02X}{b:02X}"
    return darkened_color


def select_data(df, cols=None, begin=None, end=None):
    if cols is None:
        cols = df.columns

    if begin is None:
        begin = df.index[0]

    if end is None:
        end = df.index[-1]

    return df.loc[begin:end, cols]


def find_gaps(data, cols=None, timestep=None):
    """
    Find gaps in time series data. Find individual columns gap and combined gap
    for all columns.

    Parameters:
    -----------
        data (pandas.DataFrame or pandas.Series): The time series data to check
            for gaps.

        cols (list, optional): The columns to check for gaps. Defaults to None,
            in which case all columns are checked.

        timestep (str or pandas.Timedelta, optional): The time step of the
            data. Can be either a string representation of a time period
            (e.g., '1H' for hourly data), or a pandas.Timedelta object.
            Defaults to None, in which case the time step is automatically
            determined.

    Raises:
    -------

        ValueError: If cols or timestep are invalid.

    Returns:
    --------
        dict: A dictionary containing the duration of the gaps for each
        specified column, as well as the overall combination of columns.
    """

    check_datetime_index(data)
    if not cols:
        cols = data.columns

    if not timestep:
        timestep = get_mean_timestep(data)

    # Aggregate in a single columns to know overall quality
    df = data.copy()
    df = ~df.isnull()
    df["combination"] = df.all(axis=1)

    # Index are added at the beginning and at the end to account for
    # missing values and each side of the dataset
    first_index = df.index[0] - (df.index[1] - df.index[0])
    last_index = df.index[-1] - (df.index[-2] - df.index[-1])

    df.loc[first_index] = np.ones(df.shape[1], dtype=bool)
    df.loc[last_index] = np.ones(df.shape[1], dtype=bool)
    df.sort_index(inplace=True)

    # Compute gaps duration
    res = {}
    for col in list(cols) + ["combination"]:
        time_der = df[col].loc[df[col]].index.to_series().diff()
        res[col] = time_der[time_der > timestep]

    return res


def gaps_describe(df_in, cols=None, timestep=None):
    res_find_gaps = find_gaps(df_in, cols, timestep)

    return pd.DataFrame({k: val.describe() for k, val in res_find_gaps.items()})


def get_mean_timestep(df):
    return df.index.to_frame().diff().mean()[0]


def add_scatter_and_gaps(
    figure, series, gap_series, color_rgb, alpha, y_min, y_max, yaxis
):
    figure.add_trace(
        go.Scattergl(
            x=series.index,
            y=series.to_numpy().flatten(),
            mode="lines+markers",
            name=series.name,
            yaxis=yaxis,
        )
    )

    for t_idx, gap in gap_series.items():
        figure.add_trace(
            go.Scattergl(
                x=[t_idx - gap, t_idx - gap, t_idx, t_idx],
                y=[y_min, y_max, y_max, y_min],
                mode="none",
                fill="toself",
                showlegend=False,
                fillcolor=f"rgba({color_rgb[0]}, {color_rgb[1]},"
                f" {color_rgb[2]} , {alpha})",
                yaxis=yaxis,
            )
        )


class MeasuredDats:
    def __init__(
        self,
        data,
        category_dict=None,
        category_transformations=None,
        common_transformations=None,
        resampler_agg_methods=None,
        transformers_list=None,
        config_file_path=None,
    ):
        """
        A class for handling time-series data with missing values.
        Use scikit learn Pipelines to perform operations

        Parameters:
        -----------
        data (pandas.DataFrame): The measured data.
        category_dict (dict, optional): A dictionary mapping data categories to
            column names. Defaults to None.
        category_transformations (dict, optional): A dictionary specifying
            category-specific transformations. Defaults to None. The dictionary
            keys must match the category name. For each category, a dictionary is
            specified. The keys are the transformer name, the value is a list
            ['transformer_map_name', {Corrai transformer args}]. Use only
            transformers defined in TRANSFORMER_MAP. If necessary a specific key
            "RESAMPLE" may be provided to specify an aggregation method. Method
            must be in RESAMPLE_METHS dict. If not specified, default aggreagation
            method is np.mean. An exemple of configuration is given below

        common_transformations (dict, optional): A dictionary specifying common
            transformations. The keys are the transformer name, the value is a list
            ['transformer_map_name', {Corrai transformer args}]. Use only
            transformers defined in TRANSFORMER_MAP. An example of configuration
            is given below

        resampler_agg_methods (dict, optional): A dictionary specifying the aggrgation
            method for the categories. Method must be corrai.measure RESAMPLE_METHS
            dict keys(). If no method is provided for a category, default method
            is numpy mean. Default value for this parameter is an empty dict, meaning
            aggregation method will be numpy mean for all categories.

        transformers_list (list, optional): A list of transformer names.
            Defaults to None. A list of transformer name. The order determines the
            order of the transformers in the pipeline. Note tha resample will always
            be added at the end of the pipeline. If None, a default order will be
            specified as follows: ["CATEGORY_TRANSFORMER_1", ...,
            "CATEGORY_TRANSFORMER_n", "COMMON_TRANSFORMER_1", ...,
            "CATEGORY_TRANSFORMER_n", "RESAMPLE"]


        config_file_path (str, optional): The file path for reading a json
        configuration file. Defaults to None.


        Properties:
        -----------
            columns (list): Returns the column names of the data.

            category_trans_names (list): Returns the names of category-specific
                transformations.

            common_trans_names (list): Returns the names of common
                transformations.

        Methods:
        --------
        get_pipeline(transformers_list=None, resampling_rule=False): Creates
            and returns a data processing pipeline. Custom transformer list
            may be specified. resampling_rule add a resampler to the pipeline.

        get_corrected_data(transformers_list=None, resampling_rule=False):
            Applies the pipeline to the data and returns the corrected data.
            Custom transformer list may be specified. resampling_rule add a
            resampler to the pipeline.

        get_common_transformer(transformation): Returns a pipeline for a
            common transformation.

        get_category_transformer(transformation): Returns a pipeline for a
            category-specific transformation.

        get_resampler(rule, remainder_rule="mean"): Returns a resampler for
            data resampling.

        write_config_file(file_path): Writes the current configuration to a
            file.
        read_config_file(file_path): Reads the configuration from a file.

        add_time_series(time_series, category, category_transformations=None):
            Adds a time series to the data.

        get_missing_value_stats(self, transformers_list=None,
        resampling_rule=False):
            Returns statistics on missing values for the corresponding
            transformers_list pipeline. Number of missing values for all columns
            and corresponding % of missing values

        get_gaps_description(self, cols=None, transformers_list=None,
            resampling_rule=False, gaps_timedelta=None)
            returns statistics on gaps duration for specified columns for the
            specified transformation. The column "combination" returns "aggregated"
            gaps statistics

        plot_gaps(cols=None, begin=None, end=None,
            gaps_timestep=dt.timedelta(hours=5), title="Gaps plot",
            raw_data=False, color_rgb=(243, 132, 48),  alpha=0.5, resampling_rule=False,
            transformers_list=None):
            cols (list, optional): List of column names to plot. If not
                provided, all columns will be plotted. Default is None.
            begin (str, optional): String specifying the start date for the
                data selection. If not provided, the entire dataset will be
                used. Default is None.
            end (str, optional): String specifying the end date for the data
                selection. If not provided, the entire dataset will be used.
                Default is None.
            gaps_timestep (timedelta, optional): Minimum duration between data
                points to consider a gap. Default is 5 hours.
            title (str, optional): Title of the plot. Default is "Gaps plot".
            raw_data (bool, optional): If True, plot the raw data without gap
                correction. If False, plot the gap-corrected data.
                Default is False.
            color_rgb (tuple of int, optional): RGB color of the gaps.
                Default is (243, 132, 48).
            alpha (float, optional): Opacity of the gaps. Default is 0.5.
                resampling_rule: data resampling rule
            transformers_list: transformations order list. If None it uses default
                transformers_list

        plot(cols=None, title="Correction plot", begin=None, end=None, plot_raw=False,
            plot_corrected=True, line_corrected=True, marker_corrected=True,
            line_raw=True, marker_raw=True, resampling_rule=False,
            transformers_list=None)

            Generate a plot comparing the original and corrected values of the given
            columns over the specified time range.
            cols : list of str, optional
                The names of the columns to plot. If None (default), all
                columns are plotted.
            title : str, optional
                The title of the plot. Defaults to "Correction plot".
            begin : str or datetime-like, optional
                A string or datetime-like object specifying the start of the
                time range to plot. If None (default), plot all data.
            end : str or datetime-like, optional
                A string or datetime-like object specifying the end of the
                time range to plot. If None (default), plot all data.
            plot_raw : bool, optional
                If True, plot the raw values
            plot_corrected : bool, optional
                If True, plot the corrected values .
            line_corrected: bool, optional
                If True, plot corrected values using lines
            line_raw: bool, optional
                If True, plot raw values using lines
            marker_corrected: bool, optional
                If True, plot corrected values using markers
            marker_raw: bool, optional
                If True, plot raw values using markers
            resampling_rule: False
                If resampling rule is specified, resample corrected data using
                resampler and aggregation methods specified in category_transformers.
                It will not affect raw data
            transformers_list: list, Optional
                transformations order list. Default None uses default
                transformers_list

        Example:
        --------
        >>>my_data = MeasuredDats(
            data = raw_data,
            category_dict = {
                "temperatures": [
                    'T_Wall_Ins_1', 'T_Wall_Ins_2', 'T_Ins_Ins_1', 'T_Ins_Ins_2',
                    'T_Ins_Coat_1', 'T_Ins_Coat_2', 'T_int_1', 'T_int_2', 'T_ext',
                    'T_garde'
                ],
                "illuminance": ["Lux_CW"],
                "radiation": ["Sol_rad"]
            },
            category_transformations = {
                "temperatures": {
                    "ANOMALIES": [
                        ["drop_threshold", {"upper": 100, "lower": -20}],
                        ["drop_time_gradient", {"upper_rate": 2, "lower_rate": 0}]
                    ],
                },
                "illuminance": {
                    "ANOMALIES": [
                        ["drop_threshold", {"upper": 1000, "lower": 0}],
                    ],
                },
                "radiation": {
                    "ANOMALIES": [
                        ["drop_threshold", {"upper": 1000, "lower": 0}],
                    ],
                }
            },
            common_transformations={
                "COMMON": [
                    ["interpolate", {"method": 'linear'}],
                    ["fill_na", {"method": 'bfill'}],
                    ["fill_na", {"method": 'bfill'}]
                ]
            },
            resampler_agg_methods={
                "radiation": "sum" # Method is just an example.
            },
            transformers_list=["ANOMALIES", "COMMON"]
        )

        >>>my_data.get_corrected_data()
        """
        check_datetime_index(data)
        self.data = data

        if config_file_path is None:
            self.category_dict = category_dict
            self.category_trans = category_transformations
            self.common_trans = common_transformations
            self.transformers_list = transformers_list
            self.resampler_agg_methods = resampler_agg_methods
        else:
            self.read_config_file(config_file_path)

        if self.category_dict is None:
            self.category_dict = {"data": data.columns}

        if self.category_trans is None:
            self.category_trans = {}

        if self.common_trans is None:
            self.common_trans = {}

        if self.transformers_list is None:
            self.transformers_list = self.category_trans_names + self.common_trans_names

        if self.resampler_agg_methods is None:
            self.resampler_agg_methods = {}

    @property
    def columns(self):
        return self.data.columns

    @property
    def category_trans_names(self):
        lst = [list(val.keys()) for val in self.category_trans.values()]
        lst = sum(lst, [])
        return list(dict.fromkeys(lst))

    @property
    def common_trans_names(self):
        lst = list(self.common_trans.keys())
        return list(dict.fromkeys(lst))

    def get_missing_value_stats(self, transformers_list=None, resampling_rule=False):
        data = self.get_corrected_data(transformers_list, resampling_rule)
        return missing_values_dict(data)

    def get_gaps_description(
        self,
        cols=None,
        transformers_list=None,
        resampling_rule=False,
        gaps_timedelta=None,
    ):
        if gaps_timedelta is None:
            gaps_timedelta = get_mean_timestep(self.data)
        data = self.get_corrected_data(transformers_list, resampling_rule)
        return gaps_describe(df_in=data, cols=cols, timestep=gaps_timedelta)

    def get_pipeline(self, transformers_list=None, resampling_rule=False):
        if transformers_list is None:
            transformers_list = self.transformers_list.copy()

        if "RESAMPLER" in transformers_list and not resampling_rule:
            raise ValueError(
                "RESAMPLER is present in transformers_list but no rule"
                "have been specified. use resampling_rule argument"
            )

        if resampling_rule and "RESAMPLER" not in transformers_list:
            transformers_list += ["RESAMPLER"]

        if not transformers_list:
            obj_list = [PdIdentity()]
        else:
            obj_list = []
            for trans in transformers_list:
                if trans in self.category_trans_names:
                    obj_list.append(self.get_category_transformer(trans))
                elif trans == "RESAMPLER":
                    obj_list.append(self.get_resampler(resampling_rule))
                else:
                    obj_list.append(self.get_common_transformer(trans))

        return make_pipeline(*obj_list)

    def get_corrected_data(self, transformers_list=None, resampling_rule=False):
        pipe = self.get_pipeline(
            transformers_list=transformers_list, resampling_rule=resampling_rule
        )
        return pipe.fit_transform(self.data)

    def get_common_transformer(self, transformation):
        common_trans = self.common_trans[transformation]
        return make_pipeline(
            *[TRANSFORMER_MAP[trans[0]](**trans[1]) for trans in common_trans]
        )

    def get_category_transformer(self, transformation):
        column_config_list = []
        for data_cat, cols in self.category_dict.items():
            if data_cat in self.category_trans.keys():
                if transformation in self.category_trans[data_cat].keys():
                    transformations = self.category_trans[data_cat][transformation]
                else:
                    transformations = []
                if transformations:
                    column_config_list.append(
                        (
                            f"{transformation}_{data_cat}",
                            make_pipeline(
                                *[
                                    TRANSFORMER_MAP[trans[0]](**trans[1])
                                    for trans in transformations
                                ]
                            ),
                            cols,
                        )
                    )

        return ColumnTransformer(
            column_config_list, verbose_feature_names_out=False, remainder="passthrough"
        ).set_output(transform="pandas")

    def get_resampler(self, rule, remainder_rule="mean"):
        column_config_list = []
        for data_cat, cols in self.category_dict.items():
            try:
                method = self.resampler_agg_methods[data_cat]
                column_config_list.append((cols, RESAMPLE_METHS[method]))
            except KeyError:
                pass

        if not column_config_list:
            return ct.PdResampler(rule=rule, method=np.mean)
        else:
            return ct.PdColumnResampler(
                rule=rule,
                columns_method=column_config_list,
                remainder=RESAMPLE_METHS[remainder_rule],
            )

    def write_config_file(self, file_path):
        with open(file_path, "w", encoding="utf-8") as f:
            to_dump = {
                "category_dict": self.category_dict,
                "category_transformations": self.category_trans,
                "common_transformations": self.common_trans,
                "transformers_list": self.transformers_list,
                "resampler_agg_methods": self.resampler_agg_methods,
            }
            json.dump(to_dump, f, ensure_ascii=False, indent=4)

    def read_config_file(self, file_path):
        with open(file_path, encoding="utf-8") as f:
            config_dict = json.load(f)

        attribute_list = [
            ("category_dict", "category_dict"),
            ("category_transformations", "category_trans"),
            ("common_transformations", "common_trans"),
            ("transformers_list", "transformers_list"),
            ("resampler_agg_methods", "resampler_agg_methods"),
        ]
        for attr in attribute_list:
            try:
                setattr(self, attr[1], config_dict[attr[0]])
            except KeyError:
                setattr(self, attr[1], None)

    def add_time_series(self, time_series, category, category_transformations=None):
        check_datetime_index(time_series)
        if category_transformations is None:
            category_transformations = {}

        if category in self.category_dict.keys():
            self.category_dict[category] += list(time_series.columns)
        else:
            self.category_dict[category] = list(time_series.columns)
            self.category_trans[category] = category_transformations

        self.data = pd.concat([self.data, time_series], axis=1)

    def _get_reversed_category_dict(self, cols=None):
        if cols is None:
            cols = self.data.columns

        rev_dict = {}
        for col in cols:
            for key, name_list in self.category_dict.items():
                if col in name_list:
                    rev_dict[col] = key
        return rev_dict

    def _get_yaxis_config(self, cols):
        ax_dict = self._get_reversed_category_dict(cols=cols)

        ordered_set_cat = list(dict.fromkeys(ax_dict.values()))
        ax_map = {cat: f"y{i + 1}" for i, cat in enumerate(ordered_set_cat)}

        ax_map[list(ax_map.keys())[0]] = "y"

        ax_dict = {k: ax_map[ax_dict[k]] for k in ax_dict.keys()}

        layout_ax_dict = {}
        ax_list = list(ax_map.keys())
        layout_ax_dict["yaxis"] = {"title": ax_list[0]}
        for i, ax in enumerate(ax_list[1:]):
            layout_ax_dict[f"yaxis{i + 2}"] = {"title": ax, "side": "right"}

        return ax_dict, layout_ax_dict

    def plot_gaps(
        self,
        cols=None,
        begin=None,
        end=None,
        gaps_timestep=None,
        title="Gaps plot",
        plot_raw=False,
        color_rgb=(100, 100, 100),
        alpha=0.5,
        resampling_rule=False,
        transformers_list=None,
        axis_space=0.03,
    ):
        if cols is None:
            cols = self.columns

        if gaps_timestep is None:
            gaps_timestep = dt.timedelta(hours=5)

        if plot_raw:
            to_plot = select_data(self.data, cols, begin, end)
        else:
            to_plot = select_data(
                self.get_corrected_data(
                    transformers_list=transformers_list, resampling_rule=resampling_rule
                ),
                cols,
                begin,
                end,
            )

        reversed_data_type = self._get_reversed_category_dict(cols)
        cols_data_type = defaultdict(list)
        for key, value in reversed_data_type.items():
            cols_data_type[value].append(key)
        cols_data_type = dict(cols_data_type)

        ax_dict, layout_ax_dict = self._get_yaxis_config(cols)

        fig = go.Figure()

        for col in cols:
            y_min = to_plot[cols_data_type[reversed_data_type[col]]].min().min()
            y_max = to_plot[cols_data_type[reversed_data_type[col]]].max().max()

            add_scatter_and_gaps(
                figure=fig,
                series=to_plot[col],
                gap_series=find_gaps(data=to_plot, cols=[col], timestep=gaps_timestep)[
                    col
                ],
                color_rgb=color_rgb,
                alpha=alpha,
                y_min=y_min,
                y_max=y_max,
                yaxis=ax_dict[col],
            )

        fig.update_layout(**layout_ax_dict)
        fig.update_layout(
            dict(
                title=title,
            )
        )
        fig.update_layout(
            legend=dict(
                orientation="h", yanchor="top", y=-0.1, xanchor="center", x=0.5
            ),
        )
        set_multi_yaxis_layout(figure=fig, ax_dict=ax_dict, axis_space=axis_space)
        fig.show()

    def plot(
        self,
        cols=None,
        title="Correction plot",
        begin=None,
        end=None,
        plot_raw=False,
        plot_corrected=True,
        line_corrected=True,
        marker_corrected=True,
        line_raw=True,
        marker_raw=True,
        resampling_rule=False,
        transformers_list=None,
        axis_space=0.03,
    ):
        if cols is None:
            cols = self.columns

        to_plot_raw = select_data(self.data, cols, begin, end)
        to_plot_corr = select_data(
            self.get_corrected_data(
                transformers_list=transformers_list, resampling_rule=resampling_rule
            ),
            cols,
            begin,
            end,
        )

        ax_dict, layout_ax_dict = self._get_yaxis_config(cols)

        fig = go.Figure()

        num_cols = len(cols)

        for i, col in enumerate(cols):
            if i == 0:
                # Use the first color in the palette for the first column
                color = COLOR_PALETTE[0]
            elif i == 1:
                # Use the last color in the palette for the second column
                color = COLOR_PALETTE[-1]
            elif num_cols <= 5:
                # Use the specified colors for up to 5 columns
                color = COLOR_PALETTE[i % len(COLOR_PALETTE)]
            else:
                # Generate interpolated colors for more than 5 columns
                t = (i - 2) / (num_cols - 3)  # Interpolation parameter
                color = interpolate_color(COLOR_PALETTE[0], COLOR_PALETTE[-1], t)

            dark_color = darken_color(color, 0.7)

            if line_corrected and not marker_corrected:
                mode_corrected = "lines"
            elif line_corrected and marker_corrected:
                mode_corrected = "lines+markers"
            else:
                mode_corrected = "markers"

            if plot_corrected:
                fig.add_scattergl(
                    x=to_plot_corr.index,
                    y=to_plot_corr[col],
                    name=f"{col}_corrected",
                    mode=mode_corrected,
                    line=dict(width=2, color=dark_color),
                    marker=dict(color=dark_color, opacity=0.2),
                    yaxis=ax_dict[col],
                )

            if line_raw and not marker_raw:
                mode_raw = "lines"
            elif line_raw and marker_raw:
                mode_raw = "lines+markers"
            else:
                mode_raw = "markers"

            if plot_raw:
                fig.add_scattergl(
                    x=to_plot_raw.index,
                    y=to_plot_raw[col],
                    name=f"{col}_raw",
                    mode=mode_raw,
                    marker=dict(color=color, opacity=0.5),
                    yaxis=ax_dict[col],
                )

        fig.update_layout(**layout_ax_dict)
        fig.update_layout(dict(title=title))
        fig.update_layout(
            legend=dict(
                orientation="h", yanchor="top", y=-0.1, xanchor="center", x=0.5
            ),
        )
        set_multi_yaxis_layout(figure=fig, ax_dict=ax_dict, axis_space=axis_space)
        fig.show()



import numpy as np
import pandas as pd
from pymoo.core.problem import ElementwiseProblem
from pymoo.core.variable import Integer, Real, Choice, Binary
import plotly.graph_objs as go


# TODO Add permutation variables to MyMixedProblem
class MyProblem(ElementwiseProblem):
    """
    A class that represents a single-objective optimization problem
    with one or more constraints. This class inherits from the PyMOO
    ElementwiseProblem class and overrides the _evaluate method.

    Parameters
    ----------
    parameters (list):
        A list of dictionaries, where each dictionary represents a
        parameter and contains its name, interval, and type.
    obj_func_list (list):
        A list of objects that have methods that represent the objective
        functions to be optimized or constraints (example: Modelica,
        EnergyPlus models...). If optimization is subject to inequality
        constraints g, they should be formulated as "less than" constraints
        (i.e., g(x) < 0). Leave empy if none.
    func_list (list):
        A list of objects that have methods that represent the objective
        functions to be optimized or constraints. If optimization is subject
        to inequality constraints g, they should be formulated as "less than"
        constraints (i.e., g(x) < 0). Leave empy if none.
    function_names (list):
        A list of strings that represent the names of the objective functions.
    constraint_names (list):
        A list of strings that represent the names of the constraint. Leave
        empy if none.

    Attributes
    ----------
    n_var (int):
        The number of variables inherited from parameters.
    n_obj (int):
        The number of objective functions inherited from parameters.
    n_ieq_constr (int):
        The number of constraints inherited from parameters.
    xl (numpy.ndarray):
        An array of lower bound of each variable inherited from parameters.
    xu (numpy.ndarray):
        An array of upper bound of each variable inherited from parameters.

    Methods
    -------
    _evaluate(x, out, *args, **kwargs):
        Evaluates the problem for the given variable values and returns
        the objective values out["F"] as a list of NumPy array with length of
        n_obj and the constraints values out["G"] with length of n_ieq_constr
        (if the problem has constraints to be satisfied at all).

    """

    def __init__(
        self,
        parameters,
        obj_func_list,
        func_list,
        function_names,
        constraint_names,
    ):
        self.parameters = parameters
        if len(obj_func_list) == 0 and len(func_list) == 0:
            raise ValueError(
                "At least one of obj_func_list or func_list should be provided"
            )
        self.obj_func_list = obj_func_list
        self.func_list = func_list
        self.function_names = function_names
        self.constraint_names = constraint_names

        super().__init__(
            n_var=len(parameters),
            n_obj=len(function_names),
            n_ieq_constr=len(constraint_names),
            xl=np.array([p["interval"][0] for p in parameters]),
            xu=np.array([p["interval"][-1] for p in parameters]),
        )

    def _evaluate(self, x, out, *args, **kwargs):
        current_param = {param["name"]: val for param, val in zip(self.parameters, x)}
        res = pd.concat(
            [m.function(current_param) for m in self.obj_func_list]
            + [pyf(current_param) for pyf in self.func_list]
        )
        out["F"] = list(res[self.function_names])
        out["G"] = list(res[self.constraint_names])

    def plot_pcp(self, res, ref, bounds=False):
        data_dict = {
            param["name"]: res.X[:, i]
            for param, i in zip(self.parameters, range(res.X.shape[1]))
        }
        data_dict.update(
            {self.function_names[i]: res.F[:, i] for i in range(res.F.shape[1])}
        )

        plot_parcoord(
            data_dict=data_dict,
            bounds=bounds,
            parameters=self.parameters,
            colorby=ref,
            obj_res=res.F,
        )


class MyMixedProblem(ElementwiseProblem):
    """
     A class that represents a mixed-integer optimization problem with one
     or more constraints. This class inherits from the PyMOO ElementwiseProblem
     class and overrides the _evaluate method.

     Parameters
     ----------
     parameters (list):
         A list of dictionaries, where each dictionary represents a parameter
         and contains its name, interval, and type.
     obj_func_list (list):
         A list of objects that have methods that represent the objective
         functions to be optimized. Leave empy if none
     func_list (list):
         A list of PyFunc objects that represent the constraints to be satisfied.
         Leave empy if none.
     function_names (list):
         A list of strings that represent the names of the objective functions.
         Leave empy if none.
     constraint_names (list):
         A list of strings that represent the names of the constraints.
         Leave empy if none.

     Attributes
     ----------
     vars (dict):
         A dictionary of decision variables and their types inherited
         from parameters. The variable types can be any of the following:
         - Real: continuous variable that takes values within a range of real numbers.
         - Integer: variable that takes integer values within a range.
         - Binary: variable that takes binary values (0 or 1).
         - Choice: variable that represents multiple choices from a
         set of discrete values.
     n_var (int):
        The number of variables inherited from parameters.
     n_obj (int):
        The number of objective functions inherited from parameters.
     n_ieq_constr (int):
        The number of constraints inherited from parameters.

     Methods
     -------
    _evaluate(x, out, *args, **kwargs):
        Evaluates the problem for the given variable values and returns
        the objective values out["F"] as a list of NumPy array with length of
        n_obj and the constraints values out["G"] with length of n_ieq_constr
        (if the problem has constraints to be satisfied at all).
    """

    def __init__(
        self,
        parameters,
        obj_func_list,
        func_list,
        function_names,
        constraint_names,
    ):
        self.parameters = parameters
        if len(obj_func_list) == 0 and len(func_list) == 0:
            raise ValueError(
                "At least one of obj_func_list or func_list should be provided"
            )
        self.obj_func_list = obj_func_list
        self.func_list = func_list
        self.function_names = function_names
        self.constraint_names = constraint_names
        variable_string = {}

        for param in parameters:
            name, bounds, vtype = param["name"], param["interval"], param["type"]
            if vtype == "Integer":
                var = Integer(bounds=bounds)
            elif vtype == "Real":
                var = Real(bounds=bounds)
            elif vtype == "Choice":
                var = Choice(options=bounds)
            elif vtype == "Binary":
                var = Binary()
            else:
                raise ValueError(
                    "Unknown vtype. Choose one of 'Integer', 'Real', 'Choice', "
                    "'Binary'"
                )
            variable_string[name] = var

        super().__init__(
            vars=variable_string,
            n_var=len(parameters),
            n_obj=len(function_names),
            n_ieq_constr=len(constraint_names),
        )

    def _evaluate(self, X, out, *args, **kwargs):
        res = pd.concat(
            [m.function(X) for m in self.obj_func_list]
            + [pyf(X) for pyf in self.func_list]
        )

        out["F"] = list(res[self.function_names])
        out["G"] = list(res[self.constraint_names])

    def plot_pcp(self, res, ref, bounds=False):
        data_dict = {
            param["name"]: res.X[:, i]
            for param, i in zip(self.parameters, range(res.X.shape[1]))
        }
        data_dict.update(
            {self.function_names[i]: res.F[:, i] for i in range(res.F.shape[1])}
        )

        plot_parcoord(
            data_dict=data_dict,
            bounds=bounds,
            parameters=self.parameters,
            colorby=ref,
            obj_res=res.F,
        )


def plot_parcoord(data_dict, bounds, parameters, obj_res, colorby=None):
    # Define the color palette
    color_palette = ["#FFAD85", "#FF8D70", "#ED665A", "#52E0B6", "#479A91"]

    if bounds:
        range_down = [p["interval"][0] for i, p in enumerate(parameters)]
        range_down.extend(obj_res.min(axis=0))
        range_up = [p["interval"][-1] for i, p in enumerate(parameters)]
        range_up.extend(obj_res.max(axis=0))
        ranges = [[range_down[i], range_up[i]] for i in range(len(range_up))]

    else:
        ranges = [
            [data_dict[par].min(), data_dict[par].max()] for par in data_dict.keys()
        ]

    fig = go.Figure(
        data=go.Parcoords(
            line=dict(
                color=data_dict[colorby],
                colorscale=color_palette,
                showscale=True,
                cmin=data_dict[colorby].min(),
                cmax=data_dict[colorby].max(),
            ),
            dimensions=[
                {
                    "range": r,
                    "label": par,
                    "values": data_dict[par],
                }
                for par, r in zip(data_dict.keys(), ranges)
            ],
        )
    )

    fig.show()



import pandas as pd
import numpy as np

from sklearn.base import BaseEstimator, ClusterMixin
from sklearn.neighbors import KernelDensity
from sklearn.preprocessing import StandardScaler
from scipy.signal import argrelextrema

import itertools
import datetime as dt

import plotly.colors as colors
import plotly.graph_objects as go

import corrai.custom_transformers as ct
from corrai.utils import as_1_column_dataframe
from corrai.utils import _reshape_1d
from corrai.utils import check_datetime_index
from corrai.utils import float_to_hour

from sklearn.pipeline import make_pipeline


def get_hours_switch(X, diff_filter_threshold=0, switch="positive"):
    """
    From a time series determine the number of hour since the beginning
    of the day when the signal rises (switch='positive') or decreases
    (switch='negative') or both (switch='both').
    diff_filter_threshold is used to filter small variations

    :param X: pandas Series or one column DataFrame with DatetimeIndex
    :param diff_filter_threshold: float or integer
    :param switch: 'positive', 'negative', 'both
    :return: list of hour string formatted "%H:%M"
    """

    X = as_1_column_dataframe(X)
    check_datetime_index(X)

    df = X.dropna().copy().diff()
    data_col_name = X.columns[0]
    if switch == "positive":
        df = df[df > 0]
    elif switch == "negative":
        df = abs(df[df < 0])
    elif switch == "both":
        df = abs(df)
    else:
        raise ValueError(f"Unknown value {switch} fo switch argument")

    df["start_day"] = pd.DatetimeIndex(list(map(lambda x: x.date(), df.index)))
    if df.index.tz:
        df["start_day"] = pd.DatetimeIndex(list(df["start_day"])).tz_localize(
            tz=df.index.tz
        )
    df["hour_since_beg_day"] = (df.index - df.start_day).dt.total_seconds() / 60 / 60
    df.loc[~(df[data_col_name] > diff_filter_threshold), "hour_since_beg_day"] = np.nan

    return float_to_hour(list(df.hour_since_beg_day.dropna()))


class KdeSetPointIdentificator(BaseEstimator, ClusterMixin):
    """
    KDE-based set point detection algorithm. The algorithm fits a Kernel Density
    Estimate to the provided data and detects local maxima as set points. The
    set points are then used as the basis for the clustering of new data points.
    Clustering of new points is performed by labeling each new point with the
    index of the nearest set point.

    Parameters
    ----------
    bandwidth : float, default=0.1
        The bandwidth parameter for the KDE estimator.

    domain_tol : float, default=0.2
        The tolerance parameter used for determining the range of values
        to consider when estimating the probability density function.

    domain_n_sample : int, default=1000
        The number of samples to use when estimating the probability density
        function.

    lik_filter : float, default=1
        The minimum likelihood threshold required for a set point to be
        included in the model.

    cluster_tol : float, default=0.05
        The tolerance parameter used for assigning data points to clusters.

    Attributes
    ----------
    kde : sklearn.neighbors.KernelDensity
        The KDE estimator used to estimate the probability density function.

    domain : ndarray of shape (n_samples,)
        The array of samples used to estimate the probability density function.

    set_points : ndarray of shape (n_set_points,)
        The set points detected by the algorithm.

    set_points_likelihood : ndarray of shape (n_set_points,)
        The likelihoods associated with each set point.

    Methods
    -------
    fit(X, y=None)
        Fit the model to the given data.

    predict(X)
        Use the fitted model to cluster new data points.
    """

    def __init__(
        self,
        bandwidth=0.1,
        domain_tol=0.2,
        domain_n_sample=1000,
        lik_filter=1,
        cluster_tol=0.05,
    ):
        self.bandwidth = bandwidth
        self.domain_tol = domain_tol
        self.domain_n_sample = domain_n_sample
        self.lik_filter = lik_filter
        self.cluster_tol = cluster_tol
        self.kde = KernelDensity(bandwidth=bandwidth)
        self.domain = None
        self.set_points = None
        self.set_points_likelihood = None
        self.labels_ = None

    def fit(self, X, y=None):
        X = as_1_column_dataframe(X)

        tolerance = np.mean(np.abs(X), axis=0) * self.domain_tol

        self.domain = np.linspace(
            np.min(X, axis=0) - tolerance,
            np.max(X, axis=0) + tolerance,
            self.domain_n_sample,
        )

        self.kde.fit(X.to_numpy())

        like_domain = np.exp(self.kde.score_samples(self.domain))
        max_index = argrelextrema(like_domain, np.greater)
        self.set_points = self.domain[max_index[0]]

        # Pass if perfectly flat (all values equal in the domain)
        if self.set_points.shape[0] > 0:
            set_points_likelihood = np.exp(self.kde.score_samples(self.set_points))
            like_mask = set_points_likelihood > self.lik_filter

            self.set_points = self.set_points[like_mask].flatten()
            self.set_points_likelihood = set_points_likelihood[like_mask]
            self.labels_ = self.predict(X)

        return self

    def predict(self, X):
        X = as_1_column_dataframe(X)
        X = X.to_numpy()
        X = _reshape_1d(X)

        x_cluster = np.empty(X.shape[0])
        x_cluster[:] = np.nan
        for nb, sp in enumerate(self.set_points):
            mask = np.logical_and(
                (X > (sp - self.cluster_tol)), (X < (sp + self.cluster_tol))
            )
            x_cluster[mask] = nb

        return np.nan_to_num(x_cluster, nan=-1)


def set_point_identifier(X, estimator=None, sk_scaler=None, cols=None):
    """
    Identifies set points in a time series data using kernel density estimation.
    Uses CorrAI KdeSetPointIdentificator combined with a transformer to scale the data.
    If no scaler is provided, the function uses scikit learn StandardScaler

    Parameters
    ----------
    X : pandas.DataFrame
        The input data containing time series data to identify set points in. The data
        must have a DateTimeIndex.
    estimator : object, optional
        A corrai KdeSetPointIdentificator. Default is None, which will use a
        `KdeSetPointIdentificator` object with default parameters values.
    sk_scaler : object, optional
        A scikit-learn scaler object that can transform data.
        Default is None, which will use a `StandardScaler` object wrapped in a
        corrai `PdSkTransformer` object.
    cols : list, optional
        The column names to identify set points in. Default is None, which will
        identify set points in all columns.

    Returns
    -------
    pandas.DataFrame or None
        A DataFrame containing the identified set points for each specified column,
        with MultiIndex row labels indicating the period and the set point number.
        Returns None if no set points were identified.

    Raises
    ------
    ValueError
        If the input data is not a DataFrame and does not have a DateTimeIndex.

    Notes
    -----
    The set point identification is performed by fitting the kernel density estimator
    to each column of the input data, and then finding the peaks of the density
    estimate, which shall correspond to the set points.
    """
    check_datetime_index(X)

    if isinstance(X, pd.Series):
        X = X.to_frame()

    if cols is None:
        cols = X.columns

    if sk_scaler is None:
        pd_scaler = ct.PdSkTransformer(StandardScaler())
    else:
        pd_scaler = ct.PdSkTransformer(sk_scaler)

    if estimator is None:
        estimator = KdeSetPointIdentificator()

    model = make_pipeline(pd_scaler, estimator)
    pd_scaler = model.named_steps["pdsktransformer"]
    kde = model.named_steps["kdesetpointidentificator"]

    duration = X.index[-1] - X.index[-0]

    period = pd.Period(
        X.index[0].strftime("%Y-%m-%d %H-%M-%S"),
        day=duration.days,
        second=duration.seconds,
    )

    multi_series_list = []
    for col in cols:
        model.fit(X[[col]])
        try:
            set_points = pd_scaler.inverse_transform(kde.set_points)
        except ValueError:
            set_points = None

        if set_points is not None:
            index_tuple = [
                (period, f"set_point_{i}") for i in range(set_points.shape[0])
            ]
            multiindex = pd.MultiIndex.from_tuples(index_tuple)
            set_points_series = pd.Series(
                set_points.to_numpy().flatten(), index=multiindex, name=col
            )
            multi_series_list.append(set_points_series)

    if multi_series_list:
        return pd.concat(multi_series_list, axis=1)
    else:
        return None


def moving_window_set_point_identifier(
    X, window_size, slide_size, estimator=None, sk_scaler=None, cols=None
):
    if estimator is None:
        estimator = KdeSetPointIdentificator()
    if not isinstance(estimator, KdeSetPointIdentificator):
        raise ValueError(
            f"estimator must be a corrai KdeSetPointIdentificator object "
            f"got {type(estimator)}"
        )
    if not isinstance(window_size, dt.timedelta):
        raise ValueError("window_size must be a datetime.timedelta object")
    if not isinstance(slide_size, dt.timedelta):
        raise ValueError("window_size must be a datetime.timedelta object")
    if not isinstance(X, pd.DataFrame):
        raise ValueError("data must be a Pandas DataFrame")
    if not isinstance(X.index, pd.DatetimeIndex):
        raise ValueError("data index be a Pandas DateTimeIndex")

    if cols is None:
        cols = X.columns

    start_date = X.index[0]
    end_date = X.index[-1]

    groups_res_list = []

    while start_date <= end_date - window_size:
        selected_data = X.loc[start_date : start_date + window_size, cols]
        groups_res_list.append(
            set_point_identifier(
                X=selected_data, estimator=estimator, sk_scaler=sk_scaler
            )
        )

        start_date += slide_size

    if not all(v is None for v in groups_res_list):
        return pd.concat(groups_res_list)
    else:
        return None


def plot_kde_set_point(
    X, estimator=None, sk_scaler=None, title="Clustered Timeseries", y_label="[-]"
):
    """
    Plots a scatter plot of the input data with different colors representing the
    clusters of the data points identified by the `KdeSetPointIdentificator` estimator.

    Parameters
    ----------
        X : pandas.DataFrame
            The input data with shape (n_samples, 1).
        estimator : corrai.learning.KdeSetPointIdentificator, optional
            An instance of KdeSetPointIdentificator to use for clustering the data.
            Defaults to `KdeSetPointIdentificator` with default values.
        sk_scaler : object, optional
            An instance of the scaler to use for preprocessing the
            data. Defaults to `StandardScaler`.
        title : str, optional
            The title of the plot. Defaults to "Clustered Timeseries".
        y_label : str
            The label of the y-axis. Defaults to "[-]".
    """
    X = as_1_column_dataframe(X)

    if sk_scaler is None:
        pd_scaler = ct.PdSkTransformer(StandardScaler())
    else:
        pd_scaler = ct.PdSkTransformer(sk_scaler)

    if estimator is None:
        estimator = KdeSetPointIdentificator()

    model = make_pipeline(pd_scaler, estimator)
    cluster_col = model.fit_predict(X)

    color_dict = {
        key: value
        for key, value in zip(
            sorted(set(cluster_col)), itertools.cycle(colors.DEFAULT_PLOTLY_COLORS)
        )
    }

    color_list = [color_dict[val] for val in cluster_col]

    fig = go.Figure()
    fig.add_trace(
        go.Scattergl(
            x=X.index,
            y=X.squeeze(),
            mode="markers",
            marker=dict(color=color_list),
            name=X.squeeze().name,
        )
    )

    fig.update_layout(
        dict(
            title=title,
            yaxis_title=y_label,
        )
    )

    annotation_text = (
        f"<span style='color:{list(color_dict.values())[0]}'>"
        f"Not regulated / transient</span><br>"
    )
    for i, color in enumerate(list(color_dict.values())[1:]):
        annotation_text += f"<span style='color:{color}'>" f"Set point {i}</span><br>"
    fig.add_annotation(
        xref="paper",
        yref="paper",
        x=1,
        y=1,
        text=annotation_text,
        showarrow=False,
        align="right",
        xanchor="right",
        yanchor="top",
        font=dict(size=14),
    )

    fig.show()


def plot_time_series_kde(
    X, title="Likelihood and data", x_label="", scaled=True, bandwidth=0.1, xbins=100
):
    """
    Plots the likelihood function and histogram of the input data as estimated by
    kernel density estimation.

    Parameters
    ----------
        X (pandas.DataFrame): The input data with shape (n_samples, n_features).
        title (str): The title of the plot. Defaults to "Likelihood and data".
        x_label (str): The label of the x-axis. Defaults to "".
        scaled (bool): Whether to scale the input data using `StandardScaler`.
            Defaults to True.
        bandwidth (float): The bandwidth parameter for the kernel density estimator.
            Defaults to 0.1.
        xbins (int): The number of bins to use for the histogram. Defaults to 100.
    """

    X = as_1_column_dataframe(X)

    X = X.dropna()

    scaler = StandardScaler()
    if scaled:
        X = scaler.fit_transform(X)
    else:
        X = X.to_numpy()

    domain = np.linspace(
        np.min(X, axis=0) - (np.mean(np.abs(X), axis=0) * 0.15),
        np.max(X, axis=0) + (np.mean(np.abs(X), axis=0) * 0.15),
        1000,
    )

    kde = KernelDensity(bandwidth=bandwidth)
    kde.fit(X)

    log_like_domain = np.exp(kde.score_samples(domain))

    fig = go.Figure()
    fig.add_trace(
        go.Scatter(
            x=domain.flatten(),
            y=log_like_domain.flatten(),
            mode="lines",
            name="likelihood x original",
        )
    )
    fig.add_trace(
        go.Histogram(
            x=X.squeeze(),
            histnorm="probability density",
            name="measured data",
            nbinsx=xbins,
        )
    )
    fig.update_layout(
        dict(
            title=title,
            xaxis_title=x_label,
            yaxis_title="[-]",
        )
    )
    fig.update_layout(
        legend=dict(orientation="h", yanchor="bottom", y=1, xanchor="center", x=0.5),
    )

    fig.show()



import pandas as pd
import numpy as np
from numpy.random import MT19937
from numpy.random import RandomState, SeedSequence

coefficients_COSTIC = {
    "month": {
        "January": 1.11,
        "February": 1.20,
        "March": 1.11,
        "April": 1.06,
        "May": 1.03,
        "June": 0.93,
        "July": 0.84,
        "August": 0.72,
        "September": 0.92,
        "October": 1.03,
        "November": 1.04,
        "December": 1.01,
    },
    "week": {
        "Monday": 0.97,
        "Tuesday": 0.95,
        "Wednesday": 1.00,
        "Thursday": 0.97,
        "Friday": 0.96,
        "Saturday": 1.02,
        "Sunday": 1.13,
    },
    "day": {
        "hour_weekday": np.array(
            [
                0.264,
                0.096,
                0.048,
                0.024,
                0.144,
                0.384,
                1.152,
                2.064,
                1.176,
                1.08,
                1.248,
                1.224,
                1.296,
                1.104,
                0.84,
                0.768,
                0.768,
                1.104,
                1.632,
                2.088,
                2.232,
                1.608,
                1.032,
                0.624,
            ]
        ),
        "hour_saturday": np.array(
            [
                0.408,
                0.192,
                0.072,
                0.048,
                0.072,
                0.168,
                0.312,
                0.624,
                1.08,
                1.584,
                1.872,
                1.992,
                1.92,
                1.704,
                1.536,
                1.2,
                1.248,
                1.128,
                1.296,
                1.32,
                1.392,
                1.2,
                0.936,
                0.696,
            ]
        ),
        "hour_sunday": np.array(
            [
                0.384,
                0.168,
                0.096,
                0.048,
                0.048,
                0.048,
                0.12,
                0.216,
                0.576,
                1.128,
                1.536,
                1.752,
                1.896,
                1.872,
                1.656,
                1.296,
                1.272,
                1.248,
                1.776,
                2.016,
                2.04,
                1.392,
                0.864,
                0.552,
            ]
        ),
    },
}
coefficients_RE2020 = {
    "month": {
        "January": 1.05,
        "February": 1.05,
        "March": 1.05,
        "April": 0.95,
        "May": 0.95,
        "June": 0.95,
        "July": 0.95,
        "August": 0.95,
        "September": 0.95,
        "October": 1.05,
        "November": 1.05,
        "December": 1.05,
    },
    "day": {
        "hour_weekday": np.array(
            [
                0,
                0,
                0,
                0,
                0,
                0,
                0,
                0.028,
                0.029,
                0,
                0,
                0,
                0,
                0,
                0,
                0,
                0,
                0.007,
                0.022,
                0.022,
                0.022,
                0.007,
                0.007,
                0.007,
            ]
        ),
        "hour_weekend": np.array(
            [
                0,
                0,
                0,
                0,
                0,
                0,
                0,
                0.028,
                0.029,
                0,
                0,
                0,
                0,
                0,
                0,
                0,
                0,
                0.011,
                0.011,
                0.029,
                0.011,
                0.0011,
                0.0011,
                0.0,
            ]
        ),
    },
}


class DHWaterConsumption:
    """
    Class for calculating domestic hot water consumption in a building based on either
    RE2020 or COSTIC coefficients.

    Parameters:
    -----------
    n_dwellings : int
        Number of dwellings in the building.
    v_per_dwelling : float, optional (default=110)
        Daily hot water consumption per dwelling in liters.
    ratio_bath_shower : float, optional (default=0.8)
        Ratio of baths compared to showers.
    t_shower : int, optional (default=7)
        Average duration of a shower in minutes.
    d_shower : float, optional (default=8)
       Shower flow rate of showers in liters per minute.
    s_moy_dwelling : float, optional (default=49.6)
        Average surface area per dwelling in square meters.
    s_tot_building : float, optional (default=2480)
        Total surface area of the building in square meters.
    method : str, optional (default="COSTIC")
        Method for calculating hot water consumption. Available options are "COSTIC" or
         "RE2020".

    Attributes:
    -----------
    df_coefficient : pd.DataFrame
        DataFrame containing the  coefficients for hot water consumption, calculated
        either on "COSTIC" or "RE2020" method.
    df_daily_sum : pd.DataFrame
        DataFrame containing the daily sum of coefficients.
    df_re2020 : pd.DataFrame
        DataFrame containing the hot water consumption for showers based on "RE2020"
        method.
    df_costic : pd.DataFrame
        DataFrame containing the hot water consumption for showers based on "COSTIC"
        method.
    df_costic_random : pd.DataFrame
        DataFrame containing a random distribution of hot water consumption for showers
        based on "COSTIC" method.
    df_all : pd.DataFrame
        DataFrame containing calculated coefficients and other calculated data.
    coefficients : dict
        Dictionary containing the coefficients for calculating hot water consumption
        based on the chosen method.
    v_used : float
        Total hot water consumption per shower in liters.
    v_liters_day : float
        Total hot water consumption per day for the entire building in liters.
    v_shower_bath_per_day : float
        Total hot water consumption per day for the entire building in liters taking
        into account the ratio_bath_shower.

    Methods:
    --------
    get_coefficient_calc_from_period(start, end):
        Calculates the coefficients for hot water consumption based on the chosen
        method within the given time period.
    costic_shower_distribution(start, end):
        Calculates the hot water consumption for showers based on "COSTIC" method
        within the given time period. The output dataframe is sampled every hour.
    costic_random_shower_distribution(start, end, optional_columns=False, seed=None):
        Calculates a random distribution of hot water consumption for showers based on
        "COSTIC" method within  the given time period. The output dataframe is sampled
        every minute.
        Set seed to None for a new random distribution each time method is run.
        Otherwise, set to integer for the same random distribution.
    re2020_shower_distribution(start, end):
        Calculates the hot water consumption for showers based on "COSTIC" method
        within the given time period. The output dataframe is sampled every hour.
    """

    def __init__(
        self,
        n_dwellings,
        v_per_dwelling=110,
        ratio_bath_shower=0.8,
        t_shower=7,
        d_shower=8,
        s_moy_dwelling=49.6,
        s_tot_building=2480,
        method="COSTIC",
    ):
        self.method = method
        self.n_dwellings = n_dwellings
        self.v_per_dwelling = v_per_dwelling
        self.ratio_bath_shower = ratio_bath_shower
        self.t_shower = t_shower
        self.d_shower = d_shower
        self.s_moy_dwelling = s_moy_dwelling
        self.s_tot_building = s_tot_building

        self.df_coefficient = None
        self.df_daily_sum = None
        self.df_re2020 = None
        self.df_costic = None
        self.df_costic_random = None
        self.df_all = None

        if self.method == "COSTIC":
            self.coefficients = coefficients_COSTIC
        elif self.method == "RE2020":
            self.coefficients = coefficients_RE2020

        self.v_used = self.t_shower * self.d_shower
        self.v_liters_day = self.n_dwellings * self.v_per_dwelling
        self.v_shower_bath_per_day = self.ratio_bath_shower * self.v_liters_day

    def get_coefficient_calc_from_period(self, start, end):
        if not pd.Timestamp(start) or not pd.Timestamp(end):
            raise ValueError("Start and end values must be valid timestamps.")

        date_index = pd.date_range(start=start, end=end, freq="H")
        self.df_coefficient = pd.DataFrame(
            data=np.zeros(date_index.shape[0]), index=date_index, columns=["coef"]
        )

        val_list = []
        if self.method == "COSTIC":
            for val in self.df_coefficient.index:
                if val.day_of_week in range(5):
                    hour_coefficients = self.coefficients["day"]["hour_weekday"]
                elif val.day_of_week == 5:
                    hour_coefficients = self.coefficients["day"]["hour_saturday"]
                else:
                    hour_coefficients = self.coefficients["day"]["hour_sunday"]

                h24 = (
                    hour_coefficients
                    * self.coefficients["month"][str(val.month_name())]
                    * self.coefficients["week"][str(val.day_name())]
                )
                val_list.append(h24[val.hour])

        elif self.method == "RE2020":
            for val in self.df_coefficient.index:
                if val.day_of_week in range(5):
                    hour_coefficients = self.coefficients["day"]["hour_weekday"]
                else:
                    hour_coefficients = self.coefficients["day"]["hour_weekend"]

                h24 = (
                    hour_coefficients
                    * self.coefficients["month"][str(val.month_name())]
                )
                val_list.append(h24[val.hour])

        self.df_coefficient["coef"] = val_list
        self.df_daily_sum = self.df_coefficient.resample("D").sum()
        self.df_daily_sum.columns = ["coef_daily_sum"]
        return self.df_coefficient

    def costic_shower_distribution(self, start, end):
        # Concatenation of coefficients and their daily sum
        self.get_coefficient_calc_from_period(start, end)
        df_co = pd.concat([self.df_coefficient, self.df_daily_sum], axis=1)
        df_co.fillna(method="ffill", inplace=True)
        df_co = df_co.dropna(axis=0)

        # Calculation of the number of showers per hour
        df_co["consoECS_COSTIC"] = (
            df_co["coef"] * self.v_shower_bath_per_day / df_co["coef_daily_sum"]
        )
        return df_co[["consoECS_COSTIC"]]

    def costic_random_shower_distribution(
        self, start=None, end=None, optional_columns=False, seed=None
    ):
        if seed is not None:
            rs = RandomState(MT19937(SeedSequence(seed)))
        else:
            rs = RandomState()

        periods = pd.date_range(start=start, end=end, freq="T")
        df_costic = self.costic_shower_distribution(start, end)
        df_costic["nb_shower"] = df_costic["consoECS_COSTIC"] / self.v_used
        df_costic["t_shower_per_hour"] = df_costic["nb_shower"] * self.t_shower
        df_costic["nb_shower_int"] = np.round(df_costic["nb_shower"]).astype(int)

        rs_dd = rs.randint(
            0, 60 - self.t_shower, (len(periods), df_costic["nb_shower_int"].max())
        )

        distribution_list = []
        for h, nb_shower in zip(rs_dd, df_costic["nb_shower_int"]):
            starts = h[:nb_shower]
            distribution = np.zeros(60)
            for start_shower in starts:
                distribution[start_shower : start_shower + self.t_shower] += 1
            distribution_list.append(distribution)

        df_costic_random = pd.DataFrame(
            data=np.concatenate(distribution_list),
            index=pd.date_range(
                df_costic["nb_shower_int"].index[0],
                freq="T",
                periods=df_costic.shape[0] * 60,
            ),
            columns=["shower_per_minute"],
        )

        df_costic_random["consoECS_COSTIC_random"] = (
            df_costic_random["shower_per_minute"] * self.v_used / self.t_shower
        )
        df_costic_random["consoECS_COSTIC_random"] = df_costic_random[
            "consoECS_COSTIC_random"
        ].astype(float)

        self.df_costic_random = df_costic_random

        if optional_columns:
            return df_costic_random
        else:
            return df_costic_random[["consoECS_COSTIC_random"]]

    def re2020_shower_distribution(self, start, end):
        periods = pd.date_range(start=start, end=end, freq="H")
        self.get_coefficient_calc_from_period(start, end)
        # N_calculation
        if self.s_moy_dwelling < 10:
            nmax = 1
        elif self.s_moy_dwelling in {10: 49.99}:
            nmax = 1.75 - 0.01875 * (50 - self.s_moy_dwelling)
        else:
            nmax = 0.035 * self.s_moy_dwelling

        n_adult = nmax * self.n_dwellings
        a = min(392, int(40 * (self.s_tot_building / self.s_moy_dwelling)))
        v_weekly = a * n_adult  # Liters
        v_shower_bath = v_weekly * self.ratio_bath_shower

        # Calculation of the number of showers per hour
        to_return = self.df_coefficient.copy() * v_shower_bath
        to_return = to_return[: len(periods)]
        to_return.columns = ["consoECS_RE2020"]
        return to_return


class Scheduler:
    def __init__(self, config_dict=None):
        """
        Scheduler class for generating a schedules in a DataFrame object based on
        configuration settings.

        :param dict config_dict: A dictionary containing the configuration settings
            for the scheduler. See bellow for a dictionary example.

        Guidelines for config_dict:
            - The schedules hour means "until". The first specified hour correspond to
            the first switch. In the example below, heating is 17 until 09:15 when it
            becomes 21.
            - If the set-point is constant through the day, provide the last hour of
            the day ("23:00") and the corresponding value.
            - The Scheduler allows the specification of several schedules at once.
            Each key provided in the hour dictionary correspond to a new column. If
            a key is absent at a specific hour, the previous set-point value is
            propagated

        Example of config_dict structure:
            {
                "DAYS": {
                    "working_day": {
                        "09:15": {"heating": 17, "extraction_flow_rate": 0},
                        "18:00": {"heating": 21, "extraction_flow_rate": 3000},
                        "19:00": {"heating": 22},
                        "23:00": {"heating": 17, "extraction_flow_rate": 0},
                    },
                    "Off": {
                        "23:00": {"heating": 0, "extraction_flow_rate": 0},
                    },
                    ...
                },
                "WEEKS": {
                    "winter_week": {
                        "Monday": "working_day",
                        ...
                    },
                    "summer_week": {
                        "Monday": "Off",
                        ...
                    },
                    ...
                },
                "PERIODS": [
                    (("2009-01-01", "2009-03-31"), "winter_week"),
                    (("2009-04-01", "2009-09-30"), "summer_week"),
                    (("2009-10-01", "2009-12-31"), "winter_week"),
                ],
                "TZ": "Europe/Paris",
            }

        Guidelines for config_dict:
            - The "DAYS" key should contain a dictionary where the keys represent the
            names of the schedule. Each schedule is a dictionary with hourly settings.

            - The hour in the schedule dictionary represents the "until" time.
            The first specified hour corresponds to the first switch. In the example,
            "heating" is 17 until 09:15 when it becomes 21.

            - If a set-point is constant throughout the day, provide the last hour of
            the day ("23:00") and the corresponding value.

            - The Scheduler allows the specification of several schedules at once.
            Each key provided in the hour dictionary corresponds to a new column.
            If a key is absent at a specific hour, the previous set-point value is
            propagated.

        Note:
            The config_dict should be provided during initialization and should have
            the following structure:
            - "DAYS": A dictionary containing daily schedules with specific timings
            and values.

            - "WEEKS": A dictionary mapping weekdays to the corresponding daily
            schedules.

            - "PERIODS": A list of tuples representing date periods and corresponding
            week schedules.

            - "TZ": The timezone for the scheduler.

        Example:
            scheduler = Scheduler(config_dict)
            df = scheduler.get_dataframe()

        """

        self.config_dict = config_dict

    def _get_day_dict(self):
        """
        Returns a dictionary with daily schedules based on the configuration settings.
        :return: dict containing DatFrame
        """
        day_dict = {}
        for day in self.config_dict["DAYS"].keys():
            day_df = pd.DataFrame(index=["00:00"])
            for hour, prog in self.config_dict["DAYS"][day].items():
                day_df.loc[hour, prog.keys()] = prog.values()
            day_dict[day] = day_df

        return day_dict

    def get_dataframe(self, freq="T"):
        """
        Generates and returns the scheduled DataFrame based on the configuration
        settings.

        :param str freq: output DataFrame DateTimeIndex frequency
        """

        day_list = []
        for period in self.config_dict["PERIODS"]:
            period_index = pd.date_range(
                start=period[0][0],
                end=period[0][1],
                freq="D",
                tz=self.config_dict["TZ"],
            )
            day_dict = self._get_day_dict()

            week = self.config_dict["WEEKS"][period[1]]
            for day in week.keys():
                ref_day = day_dict[week[day]]
                day_index = period_index[period_index.day_name() == day]
                for d in day_index:
                    current_day = ref_day.copy()
                    date = d.date()
                    new_index = pd.to_datetime(
                        date.strftime("%Y-%m-%d ") + current_day.index
                    )
                    current_day.index = new_index
                    day_list.append(current_day)

        df = pd.concat(day_list)
        df.sort_index(inplace=True)
        df = df.fillna(method="bfill")
        df = df.resample("T").bfill()
        df = df.shift(-1)
        df = df.fillna(method="ffill")
        return df.resample(freq).mean()


class GreyWaterConsumption:
    """
    A class to calculate the distribution of greywater
    consumption based on various factors.

    Args:
        n_people (int): Number of people in the household.
        seed (bool, optional): Whether to use a seed for random
         number generation. Defaults to False.
        dish_washer (bool, optional): Whether the household has
        a dishwasher. Defaults to True.
        washing_machine (bool, optional): Whether the household has
        a washing machine. Defaults to True.
        v_water_dish (int, optional): Volume of water used
        by the dishwasher (in liters). Defaults to 13.
        v_water_clothes (int, optional): Volume of water used by
        the washing machine (in liters). Defaults to 50.
        cycles_clothes_pers (int, optional): Number of washing machine
        cycles per person per year. Defaults to 89.
        cycles_dish_pers (int, optional): Number of dishwasher cycles
        per person per year. Defaults to 83.
        duration_dish (int, optional): Duration of a dishwasher
         cycle (in hours). Defaults to 4.
        duration_clothes (int, optional): Duration of a washing
         machine cycle (in hours). Defaults to 2.

    Raises:
        ValueError: If both dish_washer and washing_machine are False.

    Methods:
        get_GWdistribution(start, end):
            Calculates the distribution of greywater
            consumption over a given time period.

            Args:
                start (str): Start date of the time period
                (format: 'YYYY-MM-DD').
                end (str): End date of the time period
                (format: 'YYYY-MM-DD').

            Returns:
                pd.DataFrame: DataFrame containing the greywater
                consumption distribution with timestamps as index.
    """

    def __init__(
        self,
        n_people,
        seed=False,
        dish_washer=True,
        washing_machine=True,
        v_water_dish=13,
        v_water_clothes=50,
        cycles_clothes_pers=89,  # per year
        cycles_dish_pers=83,  # per year
        duration_dish=4,
        duration_clothes=2,
    ):
        self.n_people = n_people
        self.v_water_dish = v_water_dish
        self.v_water_clothes = v_water_clothes
        self.cycles_clothes_pers = cycles_clothes_pers
        self.cycles_dish_pers = cycles_dish_pers
        self.duration_dish = duration_dish
        self.duration_clothes = duration_clothes
        self.seed = seed
        self.dish_washer = dish_washer
        self.washing_machine = washing_machine

        if not dish_washer and not washing_machine:
            raise ValueError(
                "At least one of washing machine or dish washer must be True"
            )

    def get_GWdistribution(self, start, end):
        """
        Calculates the distribution of greywater consumption over a given time period.

        Args:
            start (str): Start date of the time period (format: 'YYYY-MM-DD').
            end (str): End date of the time period (format: 'YYYY-MM-DD').

        Returns:
            pd.DataFrame: DataFrame containing the greywater consumption
            distribution with timestamps as index.
        """

        date_index = pd.date_range(start=start, end=end, freq="H")
        if self.seed is not None:
            rs = RandomState(MT19937(SeedSequence(self.seed)))
        else:
            rs = RandomState()

        # calculation for dishwasher
        if self.dish_washer is True:
            Qwater_dish = self.v_water_clothes / self.duration_clothes
            dish_distribution = [0] * len(date_index)
            tot_cycles_dish_pers = int(self.cycles_dish_pers / 365 * len(date_index))

            for _ in range(self.n_people):
                k = 0  # number of cycles
                index = rs.randint(0, 120)  # time of start of cycles
                while index < (len(date_index) - 4) and k < tot_cycles_dish_pers:
                    dish_distribution[index] = dish_distribution[index] + Qwater_dish
                    dish_distribution[index + 1] = (
                        dish_distribution[index + 1] + Qwater_dish
                    )
                    dish_distribution[index + 1] = (
                        dish_distribution[index + 2] + Qwater_dish
                    )
                    dish_distribution[index + 1] = (
                        dish_distribution[index + 3] + Qwater_dish
                    )
                    space = rs.randint(72, 120)  # day 3 to day 5
                    index = index + space
                    k = k + 1

        # calculation for washing machine
        if self.washing_machine is True:
            Qwater_clothes = self.v_water_dish / self.duration_dish
            washing_distribution = [0] * len(date_index)
            tot_cycles_dish_pers = int(self.cycles_clothes_pers / 365 * len(date_index))

            for _ in range(self.n_people):
                k = 0
                index = rs.randint(0, 120)
                while index < (len(date_index) - 2) and k < tot_cycles_dish_pers:
                    washing_distribution[index] = (
                        washing_distribution[index] + Qwater_clothes
                    )
                    washing_distribution[index + 1] = washing_distribution[index + 1]
                    space = rs.randint(72, 120)
                    index = index + space
                    k = k + 1

        if self.washing_machine and self.dish_washer:
            data = [dish_distribution, washing_distribution]
            data = list(map(list, zip(*data)))
            columns = ["Q_dish", "Q_washer"]
        elif self.washing_machine and not self.dish_washer:
            data = washing_distribution
            columns = ["Q_washer"]
        else:
            data = dish_distribution
            columns = ["Q_dish"]

        df = pd.DataFrame(data=data, index=date_index, columns=columns)

        return df



import pandas as pd
import numpy as np

from scipy import integrate

from corrai.utils import as_1_column_dataframe
from corrai.utils import check_datetime_index


def time_gradient(data):
    """
    Calculates the time gradient of a given time series `data`
    between two optional time bounds `begin` and `end`.

    Parameters:
    -----------
    data : pandas Series or DataFrame
        The time series to compute the gradient on.
        If a Series is provided, it will be converted to a DataFrame
        with a single column

    Returns:
    --------
    gradient : pandas DataFrame
        A DataFrame containing the gradient of the input time series
        for each column. The index will be a DatetimeIndex
        and the columns will be the same as the input.

    Raises:
    -------
    ValueError
        If `time_series` is not a pandas Series or DataFrame.
        If the index of `time_series` is not a pandas DateTimeIndex.

    Notes:
    ------
    This function applies the `time_data_control` function to ensure that
    the input `time_series` is formatted correctly
    for time series analysis. Then, it selects a subset of the data between
     `begin` and `end` if specified. Finally, the function computes
    the gradient of each column of the subset of the data, using the
    `np.gradient` function and the time difference between consecutive
    data points.
    """

    check_datetime_index(data)
    if isinstance(data, pd.Series):
        data = as_1_column_dataframe(data)

    ts_list = []
    for col in data:
        col_ts = data[col].dropna()

        chrono = col_ts.index - col_ts.index[0]
        chrono_sec = chrono.to_series().dt.total_seconds()

        ts_list.append(
            pd.Series(np.gradient(col_ts, chrono_sec), index=col_ts.index, name=col)
        )

    return pd.concat(ts_list, axis=1)


def time_integrate(
    data, begin=None, end=None, interpolate=True, interpolation_method="linear"
):
    """
    Perform time Integration of given time series `data` between two optional
    time bounds `begin` and `end`.

    Parameters:
    -----------
    data : pandas Series or DataFrame
        The time series to integrate. If a Series is provided, it will be
        converted to a DataFrame with a single column.
    begin : str or datetime-like, optional
        Beginning time of the selection. If None, defaults to the first
        index value of `data`.
    end : str or datetime-like, optional
        End time of the selection. If None, defaults to the last index value
        of `data`.
    interpolate : bool, optional
        Whether to interpolate missing values in the input `data` before
        integrating. If True, missing values will be filled using the specified
         `interpolation_method`. If False, missing values will be
         replaced with NaNs.
    interpolation_method : str, optional
        The interpolation method to use if `interpolate` is True. Can be one
        of 'linear', 'nearest', 'zero', 'slinear', 'quadratic', 'cubic',
         or 'spline'.

    Returns:
    --------
    res_series : pandas Series
        A Series containing the result of integrating the input time series
        for each column. The index will be the same as the columns of
        the input `data`.

    Raises:
    -------
    ValueError
        If `data` is not a pandas Series or DataFrame.
        If the index of `data` is not a pandas DateTimeIndex.

    Notes:
    ------
    This function applies the `time_data_control` function to ensure that the
    input `data` is formatted correctly  for time series analysis. Then, it
    selects a subset of the data between `begin` and `end` if specified. If
    `interpolate` is True, missing values in the subset of the data will be
    filled using the specified interpolation method.
    The function then computes the integral of each column of the subset of
    the data, using the `integrate.trapz` function and the time difference
    between consecutive data points.
    """

    check_datetime_index(data)
    if isinstance(data, pd.Series):
        data = as_1_column_dataframe(data)

    if begin is None:
        begin = data.index[0]

    if end is None:
        end = data.index[-1]

    selected_ts = data.loc[begin:end, :]

    if interpolate:
        selected_ts = selected_ts.interpolate(method=interpolation_method)

    chrono = (selected_ts.index - selected_ts.index[0]).to_series()
    chrono = chrono.dt.total_seconds()

    res_series = pd.Series(dtype="float64")
    for col in data:
        res_series[col] = integrate.trapz(selected_ts[col], chrono)

    return res_series



import pandas as pd
import numpy as np
import datetime as dt
from collections.abc import Iterable


def _reshape_1d(sample):
    if isinstance(sample, pd.DataFrame):
        return sample.squeeze()
    elif isinstance(sample, np.ndarray):
        return sample.flatten()


def as_1_column_dataframe(X):
    """
    Converts a 1D array-like object to a pandas DataFrame with a single column.

    Parameters
    ----------
        X (list or numpy.ndarray or pandas.DataFrame): Input array-like object
        containing the data.

    Returns
    -------
    pandas.DataFrame: A pandas DataFrame object with a single column containing the
    input data.

    Raises
    ------
    ValueError: If input data is not a list, numpy.ndarray, or pandas.DataFrame.
    ValueError: If input data has more than one column.
    """

    if not isinstance(X, (list, np.ndarray, pd.DataFrame, pd.Series)):
        raise ValueError(
            f"X must be one of {list, np.array, pd.DataFrame}, " f"got {type(X)}"
        )

    if isinstance(X, list):
        X = pd.DataFrame(np.array(X))

    if isinstance(X, np.ndarray):
        X = pd.DataFrame(X)

    if isinstance(X, pd.Series):
        X = X.to_frame()

    if X.shape[1] > 1:
        raise ValueError(
            f"X has {X.shape[1]} columns, cannot return 1 columns DataFrame "
        )

    return X


def check_datetime_index(X):
    """Check if X is an instance od DatFrame or Series and has a DatetimeIndex"""
    if not isinstance(X, (pd.Series, pd.DataFrame)):
        raise ValueError(
            "A DataFrame or a Series was expected, got an instance of " f"{type(X)}"
        )
    if not isinstance(X.index, pd.DatetimeIndex):
        raise ValueError("X do not have a DateTimeIndex")


def apply_transformation(x, function):
    """
    Utility function to apply elementwise transformation to list, ndarray,
    Series or DataFrame. If an object other than previously mentioned is passed,
    returns function(x)
    :param x: input to be transformed
    :param function: callable
    :return: transformed list, ndarray, Series or DataFrame or object, depending on
    the input
    """
    if isinstance(x, list):
        return [function(i) for i in x]
    elif isinstance(x, np.ndarray):
        return np.vectorize(function)(x)
    elif isinstance(x, pd.Series):
        return x.apply(function)
    elif isinstance(x, pd.DataFrame):
        return x.applymap(function)

    return function(x)


def float_to_hour(hours_float):
    """
    Convert a float value representing hours to a string representation with the
    format "%H:%M".
    :param Float hours_float: Floating-point value representing hours.
    :return String representation of the hours in the format "%H:%M".
    """

    def f2s(x):
        h = int(x)
        minutes = int((x - h) * 60)
        time_str = dt.time(h, minutes).strftime("%H:%M")
        return time_str

    return apply_transformation(hours_float, f2s)


def hour_to_float(hours_string):
    """
    Convert hour string representation with the format "%H:%M"to a float value.
    :param hours_string: String, list, ndarray, Series, DataFrame value
    representing hours.
    :return float: Floating-point value representing hours.
    """

    def s2f(x):
        time_obj = dt.datetime.strptime(x, "%H:%M")
        return time_obj.hour + time_obj.minute / 60

    return apply_transformation(hours_string, s2f)


def get_reversed_dict(dictionary, values=None):
    """
    Reverses the key-value pairs in a dictionary and returns a new dictionary.

    :param dictionary: The original dictionary containing key-value pairs.

    :param values: (iterable or any) Optional. The values to filter the reversed
        dictionary by. If not provided, all values from the original dictionary will
        be used. Can be an iterable or a single value.

    :return: A new dictionary with reversed key-value pairs from the original
        dictionary. The new dictionary only includes key-value pairs where the value
        matches the specified values. If values are not provided, all key-value pairs
        from the original dictionary are included.
    """
    if values is None:
        values = dictionary.values()
    elif not isinstance(values, Iterable):
        values = [values]

    return {val: key for key, val in dictionary.items() if val in values}



import pandas as pd
import numpy as np

import pytest

from corrai.utils import _reshape_1d
from corrai.utils import as_1_column_dataframe
from corrai.utils import check_datetime_index
from corrai.utils import float_to_hour
from corrai.utils import hour_to_float
from corrai.utils import get_reversed_dict


class TestUtils:
    def test_reshape_1d(self):
        x_in = np.array([[1], [2]])
        np.testing.assert_array_equal(_reshape_1d(x_in), np.array([1, 2]))

        x_in = pd.DataFrame({"a": [1, 2]})
        pd.testing.assert_series_equal(_reshape_1d(x_in), pd.Series([1, 2], name="a"))

    def test_as_1_column_dataframe(self):
        ref = pd.DataFrame({0: [1.0, 2.0]})

        # Test list in
        pd.testing.assert_frame_equal(as_1_column_dataframe([1.0, 2.0]), ref)

        # Test DataFrame in
        pd.testing.assert_frame_equal(as_1_column_dataframe(ref), ref)

        # Test Series in
        pd.testing.assert_frame_equal(as_1_column_dataframe(ref.squeeze()), ref)

        # Test 2D array in
        pd.testing.assert_frame_equal(
            as_1_column_dataframe(np.array([[1.0], [2.0]])), ref
        )

        # Test 1D array in
        pd.testing.assert_frame_equal(as_1_column_dataframe(np.array([1.0, 2.0])), ref)

        with pytest.raises(ValueError):
            as_1_column_dataframe(1)

        with pytest.raises(ValueError):
            as_1_column_dataframe(pd.DataFrame({"a": [1], "b": [2]}))

    def test_check_datetime_index(self):
        with pytest.raises(ValueError):
            check_datetime_index(1)

        with pytest.raises(ValueError):
            check_datetime_index(pd.Series([1, 2]))

    def test_float_to_hour(self):
        assert float_to_hour(1.5) == "01:30"
        assert float_to_hour([1.5]) == ["01:30"]
        np.testing.assert_array_equal(
            float_to_hour(np.array([1.5])), np.array(["01:30"])
        )
        np.testing.assert_array_equal(
            float_to_hour(np.array([[1.5]])), np.array([["01:30"]])
        )
        pd.testing.assert_series_equal(
            float_to_hour(pd.Series([1.5])), pd.Series(["01:30"])
        )
        pd.testing.assert_frame_equal(
            float_to_hour(pd.DataFrame({"a": [2.5]})), pd.DataFrame({"a": ["02:30"]})
        )

    def test_hour_to_float(self):
        assert hour_to_float("01:30") == 1.5
        assert hour_to_float(["01:30"]) == [1.5]
        np.testing.assert_array_equal(
            hour_to_float(np.array(["01:30"])), np.array([1.5])
        )
        np.testing.assert_array_equal(
            hour_to_float(np.array([["01:30"]])), np.array([[1.5]])
        )
        pd.testing.assert_series_equal(
            hour_to_float(pd.Series(["01:30"])), pd.Series([1.5])
        )
        pd.testing.assert_frame_equal(
            hour_to_float(pd.DataFrame({"a": ["02:30"]})), pd.DataFrame({"a": [2.5]})
        )

    def test_get_reversed_dict(self):
        dictionary = {"a": 2, "b": 3, "c": 4}

        assert get_reversed_dict(dictionary, 2) == {2: "a"}
        assert get_reversed_dict(dictionary, [2, 3]) == {2: "a", 3: "b"}



import numpy as np
import pandas as pd
import datetime as dt

from scipy.ndimage import gaussian_filter1d

from sklearn.preprocessing import StandardScaler

from corrai.custom_transformers import PdDropna
from corrai.custom_transformers import PdRenameColumns
from corrai.custom_transformers import PdSkTransformer
from corrai.custom_transformers import PdDropThreshold
from corrai.custom_transformers import PdDropTimeGradient
from corrai.custom_transformers import PdApplyExpression
from corrai.custom_transformers import PdTimeGradient
from corrai.custom_transformers import PdFillNa
from corrai.custom_transformers import PdResampler
from corrai.custom_transformers import PdColumnResampler
from corrai.custom_transformers import PdAddTimeLag
from corrai.custom_transformers import PdGaussianFilter1D
from corrai.custom_transformers import PdIdentity
from corrai.custom_transformers import PdCombineColumns


class TestCustomTransformers:
    def test_pd_identity(self):
        df = pd.DataFrame({"a": [1.0]})

        identity = PdIdentity()
        res = identity.fit_transform(df)

        assert df.columns == identity.get_feature_names_out()
        pd.testing.assert_frame_equal(df, res)

    def test_pd_dropna(self):
        df = pd.DataFrame({"a": [1.0, 2.0, np.nan], "b": [3.0, 4.0, 5.0]})

        ref = pd.DataFrame({"a": [1.0, 2.0], "b": [3.0, 4.0]})

        dropper = PdDropna(how="any")

        dropper.fit(df)
        assert list(df.columns) == list(dropper.get_feature_names_out())
        pd.testing.assert_index_equal(df.index, dropper.index)
        pd.testing.assert_frame_equal(dropper.transform(df), ref)

    def test_pd_rename_columns(self):
        df = pd.DataFrame({"a": [1.0, 2.0, np.nan], "b": [3.0, 4.0, 5.0]})

        new_cols = ["c", "d"]

        renamer = PdRenameColumns(new_names=new_cols)

        renamer.fit(df)
        assert list(df.columns) == list(renamer.get_feature_names_out())
        pd.testing.assert_index_equal(df.index, renamer.index)
        assert list(renamer.transform(df).columns) == new_cols

        inversed = renamer.inverse_transform(np.zeros((2, 2)))
        assert list(inversed.columns) == new_cols

    def test_pd_sk_transformer(self):
        df = pd.DataFrame({"a": [1.0, 2.0], "b": [3.0, 4.0]})

        scaler = PdSkTransformer(StandardScaler())
        to_test = scaler.fit_transform(df)

        ref = pd.DataFrame({"a": [-1.0, 1.0], "b": [-1.0, 1.0]})

        pd.testing.assert_frame_equal(to_test, ref)
        assert list(df.columns) == list(scaler.get_feature_names_out())

        to_inverse = to_test.to_numpy()
        pd.testing.assert_frame_equal(scaler.inverse_transform(to_inverse), df)

    def test_pd_drop_threshold(self):
        df = pd.DataFrame(
            {"col1": [1, 2, 3, np.nan, 4], "col2": [1, np.nan, np.nan, 4, 5]}
        )

        ref = pd.DataFrame(
            {"col1": [np.nan, 2, 3, np.nan, 4], "col2": [np.nan, np.nan, np.nan, 4, 5]}
        )

        dropper = PdDropThreshold(lower=1.1, upper=5)
        dropper.fit(df)

        assert list(df.columns) == list(dropper.get_feature_names_out())

        pd.testing.assert_frame_equal(dropper.transform(df), ref)

        # check do nothing
        dropper = PdDropThreshold()
        pd.testing.assert_frame_equal(dropper.transform(df), df)

    def test_pd_drop_time_gradient(self):
        time_index = pd.date_range("2021-01-01 00:00:00", freq="H", periods=8)

        df = pd.DataFrame(
            {
                "dumb_column": [5, 5.1, 5.1, 6, 7, 22, 6, 5],
                "dumb_column2": [5, 5, 5.1, 6, 22, 6, np.nan, 6],
            },
            index=time_index,
        )

        ref = pd.DataFrame(
            {
                "dumb_column": [5.0, 5.1, np.nan, 6.0, 7.0, np.nan, 6.0, 5.0],
                "dumb_column2": [5.0, np.nan, 5.1, 6.0, np.nan, 6.0, np.nan, np.nan],
            },
            index=time_index,
        )

        dropper = PdDropTimeGradient(lower_rate=0, upper_rate=0.004)

        pd.testing.assert_frame_equal(ref, dropper.fit_transform(df))

        # check do nothing
        dropper = PdDropTimeGradient()
        pd.testing.assert_frame_equal(dropper.transform(df), df)

    def test_pd_apply_expression(self):
        df = pd.DataFrame({"a": [1.0, 2.0], "b": [3.0, 4.0]})

        ref = pd.DataFrame({"a": [2.0, 4.0], "b": [6.0, 8.0]})

        transformer = PdApplyExpression("X * 2")

        pd.testing.assert_frame_equal(ref, transformer.fit_transform(df))

    def test_pd_time_gradient(self):
        test = (
            pd.DataFrame(
                {"cpt1": [0, 1, 2, 2, 2, 3], "cpt2": [0, 1, 2, 2, 2, 3]},
                index=pd.date_range("2009-01-01 00:00:00", freq="10S", periods=6),
            )
            * 3600
        )

        ref = pd.DataFrame(
            {
                "cpt1": [360.0, 360.0, 180.0, -5.68e-14, 180.0, 360.0],
                "cpt2": [360.0, 360.0, 180.0, -5.68e-14, 180.0, 360.0],
            },
            index=pd.date_range("2009-01-01 00:00:00", freq="10S", periods=6),
        )

        derivator = PdTimeGradient()

        pd.testing.assert_frame_equal(ref, derivator.fit_transform(test), rtol=0.01)

    def test_pd_fill_na(self):
        test = pd.DataFrame(
            {"cpt1": [0, np.nan, 2, 2, np.nan, 3], "cpt2": [0, 1, 2, 2, np.nan, 3]}
        )

        ref = pd.DataFrame(
            {
                "cpt1": [0.0, 2.0, 2.0, 2.0, 3.0, 3.0],
                "cpt2": [0.0, 1.0, 2.0, 2.0, 3.0, 3.0],
            }
        )

        filler = PdFillNa(method="bfill")
        pd.testing.assert_frame_equal(ref, filler.fit_transform(test))

        ref = pd.DataFrame(
            {
                "cpt1": [0.0, 0.0, 2.0, 2.0, 0.0, 3.0],
                "cpt2": [0.0, 1.0, 2.0, 2.0, 0.0, 3.0],
            }
        )

        filler = PdFillNa(value=0.0)
        pd.testing.assert_frame_equal(ref, filler.fit_transform(test))

    def test_pd_resampler(self):
        test = pd.DataFrame(
            {"col": [1.0, 2.0, 3.0]},
            index=pd.date_range("2009-01-01 00:00:00", freq="H", periods=3),
        )

        ref = pd.DataFrame(
            {"col": [2.0]},
            index=pd.DatetimeIndex(["2009-01-01"], dtype="datetime64[ns]", freq="3H"),
        )

        transformer = PdResampler(rule="3H", method=np.mean)

        pd.testing.assert_frame_equal(ref, transformer.fit_transform(test))

    def test_pd_columns_resampler(self):
        np.random.seed(42)
        df = pd.DataFrame(
            {
                "col0": np.arange(10) * 100,
                "col1": np.arange(10),
                "col2": np.random.random(10),
                "col3": np.random.random(10) * 10,
            },
            index=pd.date_range("2009-01-01", freq="H", periods=10),
        ).astype("float")

        ref = pd.DataFrame(
            {
                "col0": [400.0, 900.0],
                "col1": [2.0, 7.0],
                "col2": [0.56239, 0.47789],
                "col3": [9.69910, 5.24756],
            },
            index=pd.DatetimeIndex(
                ["2009-01-01 00:00:00", "2009-01-01 05:00:00"],
                dtype="datetime64[ns]",
                freq="5H",
            ),
        ).astype("float")

        column_resampler = PdColumnResampler(
            rule="5H",
            columns_method=[(["col2"], np.mean), (["col1"], np.mean)],
            remainder=np.max,
        )

        pd.testing.assert_frame_equal(
            ref, column_resampler.fit_transform(df).astype("float"), atol=0.01
        )

        column_resampler = PdColumnResampler(
            rule="5H",
            columns_method=[(["col2"], np.mean), (["col1"], np.mean)],
            remainder="drop",
        )

        pd.testing.assert_frame_equal(
            ref[["col2", "col1"]],
            column_resampler.fit_transform(df).astype("float"),
            atol=0.01,
        )

    def test_pd_add_time_lag(self):
        df = pd.DataFrame(
            {
                "col0": np.arange(2),
                "col1": np.arange(2) * 10,
            },
            index=pd.date_range("2009-01-01", freq="H", periods=2),
        )

        ref = pd.DataFrame(
            {
                "col0": [1.0],
                "col1": [10.0],
                "1:00:00_col0": [0.0],
                "1:00:00_col1": [0.0],
            },
            index=pd.DatetimeIndex(
                ["2009-01-01 01:00:00"], dtype="datetime64[ns]", freq="H"
            ),
        )

        lager = PdAddTimeLag(time_lag=dt.timedelta(hours=1), drop_resulting_nan=True)

        pd.testing.assert_frame_equal(ref, lager.fit_transform(df))

    def test_pd_gaussian_filter(self):
        df = pd.DataFrame({"a": [1, 2, 3], "b": [4, 5, 6]})

        gfilter = PdGaussianFilter1D()

        to_test = gfilter.fit_transform(df)

        np.testing.assert_almost_equal(
            gaussian_filter1d(
                df.to_numpy()[:, 0].T, sigma=5, mode="nearest", truncate=4.0
            ),
            to_test.to_numpy()[:, 0],
            decimal=5,
        )

        assert list(to_test.columns) == list(df.columns)

    def test_pd_combine_columns(self):
        x_in = pd.DataFrame({"a": [1, 2], "b": [1, 2], "c": [1, 2]})

        trans = PdCombineColumns(
            columns_to_combine=["a", "b"],
            function=np.sum,
            function_kwargs={"axis": 1},
            drop_columns=True,
        )

        pd.testing.assert_frame_equal(
            trans.fit_transform(x_in), pd.DataFrame({"c": [1, 2], "combined": [2, 4]})
        )

        ref = x_in.copy()
        ref["combined"] = [2, 4]
        trans.drop_columns = False

        pd.testing.assert_frame_equal(trans.fit_transform(x_in), ref)



import numpy as np
import pandas as pd
import json

from corrai.measure import MeasuredDats
from corrai.measure import missing_values_dict
from corrai.measure import gaps_describe
from corrai.measure import select_data
from copy import deepcopy

from pathlib import Path

import pytest

RESOURCES_DIR = Path(__file__).parent / "resources"

TEST_DF = pd.DataFrame(
    {
        "dumb_column": [-1, 5, 100, 5, 5.1, 5.1, 6, 7, 22, 6, 5],
        "dumb_column2": [-10, 50, 1000, 50, 50.1, 50.1, 60, 70, 220, 60, 50],
        "dumb_column3": [-100, 500, 10000, 500, 500.1, 500.1, 600, 700, 2200, 600, 500],
    },
    index=pd.date_range("2021-01-01 00:00:00", freq="H", periods=11),
)


@pytest.fixture(scope="session")
def my_measure():
    tested_obj = MeasuredDats(
        data=TEST_DF,
        category_dict={
            "col_1": ["dumb_column"],
            "col_2": ["dumb_column2"],
            "col_3": ["dumb_column3"],
        },
        category_transformations={
            "col_1": {
                "ANOMALIES": [
                    ["drop_threshold", {"upper": 50, "lower": 0}],
                    ["drop_time_gradient", {"lower_rate": 0, "upper_rate": 0.004}],
                ],
                "PROCESS": [["apply_expression", {"expression": "X * 2"}]],
            },
            "col_2": {
                "ANOMALIES": [
                    ["drop_threshold", {"upper": 500, "lower": 0}],
                ],
            },
            "col_3": {},
        },
        common_transformations={
            "COMMON": [
                ["interpolate", {"method": "linear"}],
                ["fill_na", {"method": "bfill"}],
                ["fill_na", {"method": "ffill"}],
            ]
        },
        resampler_agg_methods={"col_2": "sum"},
    )

    return tested_obj


class TestMeasuredDats:
    def test_rw_config_file(self, my_measure, tmp_path_factory):
        test_save_path = tmp_path_factory.mktemp("save")

        my_measure_loc = deepcopy(my_measure)
        my_measure_loc.transformers_list = ["PROCESS", "COMMON"]
        my_measure_loc.write_config_file(test_save_path / "save.json")

        to_test = MeasuredDats(
            data=TEST_DF, config_file_path=test_save_path / "save.json"
        )

        assert to_test.category_dict == my_measure_loc.category_dict
        assert to_test.category_trans == my_measure_loc.category_trans
        assert to_test.common_trans == my_measure_loc.common_trans
        assert to_test.transformers_list == my_measure_loc.transformers_list

        with open(test_save_path / "save.json", "w", encoding="utf-8") as f:
            json.dump({}, f, ensure_ascii=False, indent=4)

        to_test = MeasuredDats(
            data=TEST_DF, config_file_path=test_save_path / "save.json"
        )

        assert to_test.category_dict == {"data": TEST_DF.columns}
        assert to_test.category_trans == {}
        assert to_test.common_trans == {}
        assert to_test.transformers_list == []

    def test_select_data(self):
        df = pd.DataFrame(
            {
                "a": [1, 2],
                "b": [1, 2],
            }
        )

        pd.testing.assert_frame_equal(select_data(df), df)

        ref = pd.DataFrame({"a": [1]})

        pd.testing.assert_frame_equal(select_data(df, begin=0, end=0, cols=["a"]), ref)

    def test_remove_anomalies(self, my_measure):
        ref_anomalies = pd.DataFrame(
            {
                "dumb_column": [
                    np.nan,
                    5,
                    np.nan,
                    np.nan,
                    5.1,
                    np.nan,
                    6,
                    7,
                    np.nan,
                    6,
                    5,
                ],
                "dumb_column2": [
                    np.nan,
                    50.0,
                    np.nan,
                    50.0,
                    50.1,
                    50.1,
                    60,
                    70,
                    220.0,
                    60,
                    50,
                ],
                "dumb_column3": [
                    -100,
                    500,
                    10000,
                    500,
                    500.1,
                    500.1,
                    600,
                    700,
                    2200,
                    600,
                    500,
                ],
            },
            index=TEST_DF.index,
        )

        assert ref_anomalies.equals(
            my_measure.get_corrected_data(transformers_list=["ANOMALIES"])
        )

        ref = pd.DataFrame(
            {
                "dumb_column": [10.08, 12.42, 10.0],
                "dumb_column2": [250.10, 460.10, 50.0],
                "dumb_column3": [2280.02000, 920.02000, 500.00000],
            },
            index=pd.date_range("2021-01-01", freq="5H", periods=3),
        )

        pd.testing.assert_frame_equal(
            my_measure.get_corrected_data(resampling_rule="5H"), ref
        )

    def test_missing_values_dict(self):
        time_index_df = pd.date_range("2021-01-01 00:00:00", freq="30T", periods=4)

        df = pd.DataFrame(
            {
                "dumb_column": [5.0, np.nan, 6.0, 6.0],
                "dumb_column2": [5.0, 5.0, np.nan, np.nan],
            },
            index=time_index_df,
        )

        res = {
            "Number_of_missing": pd.Series([3, 2], index=df.columns),
            "Percent_of_missing": pd.Series([25.0, 50.0], index=df.columns),
        }

        to_test = missing_values_dict(df)

        assert res["Number_of_missing"].equals(to_test["Number_of_missing"])
        assert res["Percent_of_missing"].equals(to_test["Percent_of_missing"])

    def test_gaps_describe(self):
        time_index_df = pd.date_range("2021-01-01 00:00:00", freq="H", periods=5)

        df = pd.DataFrame(
            {
                "dumb_column1": [np.nan, 5, 5, 5, 5],
                "dumb_column2": [5, np.nan, 5, 5, 5],
                "dumb_column3": [5, np.nan, np.nan, 5, 5],
                "dumb_column4": [5, 5, 5, 5, np.nan],
            },
            index=time_index_df,
        )

        one_hour_dt = pd.to_timedelta("2H")
        one_n_half = pd.to_timedelta("2H30min")
        two_hour_dt = pd.to_timedelta("3H")
        two_n_half = pd.to_timedelta("3H30min")
        three_hour = pd.to_timedelta("4H")
        spec_std = pd.to_timedelta("0 days 01:24:51.168824543")
        nat = pd.NaT

        ref = pd.DataFrame(
            {
                "dumb_column1": [
                    1,
                    one_hour_dt,
                    nat,
                    one_hour_dt,
                    one_hour_dt,
                    one_hour_dt,
                    one_hour_dt,
                    one_hour_dt,
                ],
                "dumb_column2": [
                    1,
                    one_hour_dt,
                    nat,
                    one_hour_dt,
                    one_hour_dt,
                    one_hour_dt,
                    one_hour_dt,
                    one_hour_dt,
                ],
                "dumb_column3": [
                    1,
                    two_hour_dt,
                    nat,
                    two_hour_dt,
                    two_hour_dt,
                    two_hour_dt,
                    two_hour_dt,
                    two_hour_dt,
                ],
                "dumb_column4": [
                    1,
                    one_hour_dt,
                    nat,
                    one_hour_dt,
                    one_hour_dt,
                    one_hour_dt,
                    one_hour_dt,
                    one_hour_dt,
                ],
                "combination": [
                    2,
                    two_hour_dt,
                    spec_std,
                    one_hour_dt,
                    one_n_half,
                    two_hour_dt,
                    two_n_half,
                    three_hour,
                ],
            },
            index=["count", "mean", "std", "min", "25%", "50%", "75%", "max"],
        )

        assert ref.equals(gaps_describe(df))

    def test_get_reversed_data_type_dict(self, my_measure):
        to_test = my_measure._get_reversed_category_dict(
            ["dumb_column", "dumb_column2"]
        )

        assert to_test == {"dumb_column": "col_1", "dumb_column2": "col_2"}

    def test_get_yaxis_config(self, my_measure):
        ax_dict, layout_ax_dict = my_measure._get_yaxis_config(
            cols=["dumb_column", "dumb_column2", "dumb_column3"]
        )

        ax_dict_ref = {"dumb_column": "y", "dumb_column2": "y2", "dumb_column3": "y3"}
        layout_ax_dict_ref = {
            "yaxis": {"title": "col_1"},
            "yaxis2": {"title": "col_2", "side": "right"},
            "yaxis3": {"title": "col_3", "side": "right"},
        }

        assert ax_dict == ax_dict_ref
        assert layout_ax_dict == layout_ax_dict_ref

    def test_add_time_series(self):
        data = pd.DataFrame(
            {"col": [1, 2, 3]},
            index=pd.date_range("2009-01-01 00:00:00", freq="H", periods=3),
        )

        mdata = MeasuredDats(
            data.copy(),
            category_dict={"dumb_type": ["col"]},
            category_transformations={"dumb_type": {}},
            common_transformations={},
        )

        new_dat = pd.DataFrame(
            {"col1": [2, 3, 4]},
            index=pd.date_range("2009-01-01 01:00:00", freq="H", periods=3),
        )

        mdata.add_time_series(new_dat, category="dumb_type")

        assert mdata.category_dict == {"dumb_type": ["col", "col1"]}
        assert mdata.category_trans == {"dumb_type": {}}

        new_dat.columns = ["col2"]

        mdata.add_time_series(new_dat, category="new_dumb_type")

        assert mdata.category_dict == {
            "dumb_type": ["col", "col1"],
            "new_dumb_type": ["col2"],
        }
        assert mdata.category_trans == {"dumb_type": {}, "new_dumb_type": {}}

        new_dat.columns = ["col3"]

        mdata.add_time_series(
            new_dat,
            category="new_dumb_type2",
            category_transformations={"resample": "mean"},
        )

        assert mdata.category_dict == dict(
            dumb_type=["col", "col1"], new_dumb_type=["col2"], new_dumb_type2=["col3"]
        )
        assert mdata.category_trans == {
            "dumb_type": {},
            "new_dumb_type": {},
            "new_dumb_type2": {"resample": "mean"},
        }

    def test_plot_gap(self, my_measure):
        import datetime as dt

        my_measure.plot_gaps(
            gaps_timestep=dt.timedelta(hours=1), transformers_list=["ANOMALIES"]
        )

    def test_plot(self, my_measure):
        my_measure.plot(
            begin="2021-01-01 02:00:00",
            plot_raw=True,
            resampling_rule="2H",
        )



import pandas as pd
import numpy as np
from sklearn.preprocessing import StandardScaler
from sklearn.pipeline import Pipeline

from corrai.learning import KdeSetPointIdentificator
from corrai.learning import get_hours_switch
from corrai.learning import plot_kde_set_point, plot_time_series_kde
from corrai.learning import set_point_identifier, moving_window_set_point_identifier
import corrai.custom_transformers as ct

import datetime as dt
from pathlib import Path

FILES_PATH = Path(__file__).parent / "resources"


class TestLearning:
    def test_kde_set_point_identificator(self):
        data = pd.read_csv(
            FILES_PATH / "kde_test_dataset.csv", index_col=0, parse_dates=True
        )

        pipe = Pipeline(
            [
                ("drop_na", ct.PdDropna()),
                ("scaler", ct.PdSkTransformer(StandardScaler())),
                ("kde_cluster", KdeSetPointIdentificator()),
            ]
        )

        pipe.fit(data)

        np.testing.assert_array_almost_equal(
            pipe.named_steps["kde_cluster"].set_points,
            np.array([-0.80092, 1.25735]),
            decimal=3,
        )

        np.testing.assert_array_almost_equal(
            pipe.named_steps["kde_cluster"].set_points_likelihood,
            np.array([2.43101, 1.49684]),
            decimal=3,
        )

        clustered = pipe.predict(pd.Series([4400, 0, 156]).to_frame())

        np.testing.assert_array_almost_equal(clustered, np.array([1, 0, -1]))

    def test_switch(self):
        test_series = pd.Series(
            [0, 0, 1000, 1020, 0.1, 0.3],
            index=pd.date_range(
                "2009-01-01 00:00:00", freq="H", periods=6, tz="Europe/Paris"
            ),
            name="flwr",
        )
        assert True

        assert get_hours_switch(test_series, 200, switch="positive") == ["02:00"]
        assert get_hours_switch(test_series, 200, switch="negative") == ["04:00"]
        assert get_hours_switch(test_series, 200, switch="both") == ["02:00", "04:00"]

    def test_plot_kde_set_point(self):
        # Generate test data
        x = pd.Series(np.random.randn(100), name="x")

        estimator = KdeSetPointIdentificator()
        estimator.fit(x.to_frame())

        # Call the function with default arguments
        plot_kde_set_point(x.to_frame(), estimator)

        # Call the function with non-default arguments
        plot_kde_set_point(
            x.to_frame(),
            estimator,
            title="Clustered Timeseries",
            y_label="test_lab",
        )

    def test_plot_time_series_kde(self):
        # Generate test data
        x = pd.Series(np.random.randn(100), name="x")

        # Call the function with default arguments
        plot_time_series_kde(x.to_frame())

        # Call the function with non-default arguments
        plot_time_series_kde(
            x.to_frame(),
            title="Likelihood and data",
            x_label="x",
            scaled=False,
            bandwidth=0.2,
            xbins=200,
        )

    def test_set_point_identifier(self):
        f_data = pd.read_csv(
            FILES_PATH / "kde_false_data.csv", index_col=0, parse_dates=True
        )

        res = set_point_identifier(
            f_data, estimator=KdeSetPointIdentificator(bandwidth=0.1, lik_filter=0.6)
        )

        ref = pd.DataFrame(
            {
                "a": {
                    (
                        pd.Period("2009-01-01 00:00", "H"),
                        "set_point_0",
                    ): 122.65772294951199,
                    (
                        pd.Period("2009-01-01 00:00", "H"),
                        "set_point_1",
                    ): 241.75480686502476,
                    (
                        pd.Period("2009-01-01 00:00", "H"),
                        "set_point_2",
                    ): 387.6906702544559,
                }
            }
        )

        pd.testing.assert_frame_equal(res, ref)

    def test_moving_window_set_point_identifier(self):
        f_data = pd.read_csv(
            FILES_PATH / "kde_false_data.csv", index_col=0, parse_dates=True
        )

        res = moving_window_set_point_identifier(
            f_data,
            window_size=dt.timedelta(hours=10),
            slide_size=dt.timedelta(hours=10),
        )

        ref = pd.DataFrame(
            {
                "a": {
                    (
                        pd.Period("2009-01-01 00:00", "H"),
                        "set_point_0",
                    ): 122.28125863629094,
                    (
                        pd.Period("2009-01-01 00:00", "H"),
                        "set_point_1",
                    ): 384.76032496023265,
                    (
                        pd.Period("2009-01-01 10:00", "H"),
                        "set_point_0",
                    ): 245.13766266266265,
                }
            }
        )

        pd.testing.assert_frame_equal(res, ref)



import numpy as np
from corrai.metrics import nmbe
from corrai.metrics import cv_rmse


class TestMetrics:
    def test_nmbe(self):
        y_true = np.array([1, 2, 3, 4])
        y_pred = np.array([1.5, 2, 2.2, 3])

        expected = -13.0

        np.testing.assert_allclose(
            nmbe(y_pred=y_pred, y_true=y_true), expected, rtol=10 - 7
        )

    def test_cv_rmse(self):
        y_true = np.array([1, 2, 3, 4])
        y_pred = np.array([1.5, 2, 2.2, 3])

        expected = 30.0

        np.testing.assert_allclose(
            cv_rmse(y_pred=y_pred, y_true=y_true), expected, rtol=10 - 7
        )



import numpy as np
from sklearn.utils import check_consistent_length
from sklearn.metrics._regression import _check_reg_targets


def nmbe(y_pred, y_true):
    """Normalized Mean Biased Error

    :param y_pred: array-like of shape (n_samples,) or (n_samples, n_outputs)
        Ground truth (correct) target values.
    :param y_true: array-like of shape (n_samples,) or (n_samples, n_outputs)
        Estimated target values.
    :return:
    Normalized Mean biased error as float
    """
    check_consistent_length(y_pred, y_true)
    y_type, y_true, y_pred, multioutput = _check_reg_targets(
        y_true, y_pred, "uniform_average"
    )

    return np.sum(y_pred - y_true) / np.sum(y_true) * 100


def cv_rmse(y_pred, y_true):
    """Coefficient of variation of root mean squared error

    :param y_pred: array-like of shape (n_samples,) or (n_samples, n_outputs)
        Ground truth (correct) target values.
    :param y_true: array-like of shape (n_samples,) or (n_samples, n_outputs)
        Estimated target values.
    :return:
    Coefficient of variation of root mean squared error as float
    """
    check_consistent_length(y_pred, y_true)
    y_type, y_true, y_pred, multioutput = _check_reg_targets(
        y_true, y_pred, "uniform_average"
    )
    return (
        (1 / np.mean(y_true))
        * np.sqrt(np.sum((y_true - y_pred) ** 2) / (y_true.shape[0] - 1))
        * 100
    )



import pandas as pd
import numpy as np
from pathlib import Path

import pytest

from corrai.multi_optimize import MyProblem, MyMixedProblem

from pymoo.algorithms.moo.nsga2 import NSGA2
from pymoo.algorithms.soo.nonconvex.ga import GA
from pymoo.operators.crossover.sbx import SBX
from pymoo.operators.mutation.pm import PM
from pymoo.operators.repair.rounding import RoundingRepair
from pymoo.operators.sampling.rnd import IntegerRandomSampling
from pymoo.algorithms.soo.nonconvex.de import DE
from pymoo.operators.sampling.lhs import LHS
from pymoo.optimize import minimize


PACKAGE_DIR = Path(__file__).parent / "TestLib"


def py_func_rosen(x_dict):
    return pd.Series(
        (1 - x_dict["x"]) ** 2 + 100 * (x_dict["y"] - x_dict["x"] ** 2) ** 2,
        index=["f1"],
    )


class MyObjectBinhandKorn1:
    def function(self, x):
        f1 = 4 * x["x"] ** 2 + 4 * x["y"] ** 2
        f2 = (x["x"] - 5) ** 2 + (x["y"] - 5) ** 2
        g1 = (x["x"] - 5) ** 2 + x["y"] ** 2 - 25
        return pd.Series([f1, f2, g1], index=["f1", "f2", "g1"])


class MyObjectBinhandKorn2:
    def function(self, x):
        g2 = 7.7 - (x["x"] - 8) ** 2 - (x["y"] + 3) ** 2
        return pd.Series([g2], index=["g2"])


class MyObjectMixed:
    def function(self, x):
        f1 = x["z"] ** 2 + x["y"] ** 2
        f2 = (x["z"] + 2) ** 2 + (x["y"] - 1) ** 2

        if x["b"]:
            f2 = 100 * f2

        if x["x"] == "multiply":
            f2 = 10 * f2

        return pd.Series([f1, f2], index=["f1", "f2"])


parameters = [
    {"name": "x", "interval": (-2, 10)},
    {"name": "y", "interval": (-2, 10)},
]


class TestMyProblem:
    def test_myproblem_simple(self):
        problem = MyProblem(
            parameters=parameters,
            obj_func_list=[],
            func_list=[py_func_rosen],
            function_names=["f1"],
            constraint_names=[],
        )

        algorithm = DE(pop_size=100, sampling=LHS(), CR=0.3, jitter=False)

        res = minimize(problem, algorithm, seed=1, verbose=False)

        np.testing.assert_allclose(res.X, np.array([1, 1]), rtol=0.01)

    def test_myproblem_twoobjectsfunction(self):
        param = [
            {"name": "x", "interval": (0, 5)},
            {"name": "y", "interval": (0, 3)},
        ]

        obj1 = MyObjectBinhandKorn1()
        obj2 = MyObjectBinhandKorn2()

        problem = MyProblem(
            parameters=param,
            obj_func_list=[obj1, obj2],
            func_list=[],
            function_names=["f1", "f2"],
            constraint_names=["g1", "g2"],
        )

        res = minimize(
            problem, NSGA2(pop_size=10), ("n_gen", 10), seed=1, verbose=False
        )

        assert np.allclose(
            res.X,
            np.array(
                [
                    [0.08320695, 0.05947538],
                    [3.11192632, 2.89802109],
                    [2.21805962, 2.98362625],
                    [2.43878853, 2.22367131],
                    [0.52530151, 0.08202677],
                    [0.76324568, 0.9977904],
                    [1.36107106, 0.97739205],
                    [1.86374614, 1.50234103],
                    [2.75558726, 2.98367642],
                    [1.68980871, 0.95713564],
                ]
            ),
        ) or np.allclose(
            res.X,
            np.array(
                [
                    [3.0416415, 2.4492151],
                    [0.3985999, 0.2485888],
                    [2.7348734, 2.0720593],
                    [0.7103709, 1.0476442],
                    [1.9838374, 1.0405508],
                    [1.7628016, 2.0258502],
                    [3.0416415, 2.4184421],
                    [1.9694239, 1.5152536],
                    [1.7935434, 2.5076411],
                    [0.8009073, 1.0476442],
                ]
            ),
        )

    def test_myproblem_integers(self):
        problem = MyProblem(
            parameters=parameters,
            obj_func_list=[],
            func_list=[py_func_rosen],
            function_names=["f1"],
            constraint_names=[],
        )

        # for  integer variables only --> no need to specify type integer
        method = GA(
            pop_size=20,
            sampling=IntegerRandomSampling(),
            crossover=SBX(prob=1.0, eta=3.0, vtype=float, repair=RoundingRepair()),
            mutation=PM(prob=1.0, eta=3.0, vtype=float, repair=RoundingRepair()),
            eliminate_duplicates=True,
        )

        res = minimize(
            problem, method, termination=("n_gen", 40), seed=1, verbose=False
        )

        assert np.allclose(res.X, np.array([1, 1]), rtol=0, atol=0)

    def test_myproblem_mixed(self):
        param = [
            {"name": "b", "interval": (), "type": "Binary"},
            {"name": "x", "interval": ("nothing", "multiply"), "type": "Choice"},
            {"name": "y", "interval": (-2, 2.5), "type": "Integer"},
            {"name": "z", "interval": (-5, 5), "type": "Real"},
        ]

        obj = MyObjectMixed()

        problem = MyMixedProblem(
            parameters=param,
            obj_func_list=[obj],
            func_list=[],
            function_names=["f1", "f2"],
            constraint_names=[],
        )

        to_test = problem.evaluate(
            [{"b": False, "x": "nothing", "y": -1, "z": -3.171895287195006}],
            return_as_dictionary=True,
        )

        ref_f = np.array([[11.060919712929891, 5.373338564149866]])
        ref_g = np.array([[]])

        assert np.array_equal(to_test["F"], ref_f)
        assert np.array_equal(to_test["G"], ref_g)

    def test_warning_error(self):
        with pytest.raises(ValueError):
            MyProblem(
                parameters=parameters,
                obj_func_list=[],
                func_list=[],
                function_names=["f1"],
                constraint_names=[],
            )

        with pytest.raises(ValueError):
            MyMixedProblem(
                parameters=parameters,
                obj_func_list=[],
                func_list=[],
                function_names=["f1"],
                constraint_names=[],
            )
